;; Copyright (c) 1999-2000 David Kuehling  <dvdkhlng TA gmx TOD de>

;; Permission is hereby granted, free of charge, to any person obtaining a copy
;; of this software and associated documentation files (the "Software"), to
;; deal in the Software without restriction, including without limitation the
;; rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
;; sell copies of the Software, and to permit persons to whom the Software is
;; furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice (including the next
;; paragraph) shall be included in all copies or substantial portions of the
;; Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
;; IN THE SOFTWARE.
; Unfortunately, this file needs to be preprocessed by my PreFargo
; preprocessor before you can compile it. Send me a mail for it or have a 
; look at http://www.ticalc.org
; If I've some time, I'll try to convert it to a normal .asm source file.
;
;

/******************************************************************************
 * Library for (but not only) FORTH92
 *-----------------------------------------------------------------------------
 *
 *
 * All the routines of this library may modify the registers d0-d2/a0-a1,
 * if nothing else is specified.
 */


/******************************************************************************/
		#include <tios.h>
/******************************************************************************/

/******************************************************************************/
		XDEF	_library
		XDEF	forthlib@0000
		XDEF	forthlib@0001
		XDEF	forthlib@0002
		XDEF	forthlib@0003
		XDEF	forthlib@0004
		XDEF	forthlib@0005
		XDEF	forthlib@0006
		XDEF	forthlib@0007
		XDEF	forthlib@0008
		XDEF	forthlib@0009
		XDEF	forthlib@000A
		XDEF	forthlib@000B
		XDEF	forthlib@000C
		XDEF	forthlib@000D
		XDEF	forthlib@000E
		XDEF	forthlib@000F
		XDEF	forthlib@0010
		XDEF	forthlib@0011
		XDEF	forthlib@0012
		XDEF	forthlib@0013
		XDEF	forthlib@0014
		XDEF	forthlib@0015
		XDEF	forthlib@0016
		XDEF	forthlib@0017
		XDEF	forthlib@0018
		XDEF	forthlib@0019
		XDEF    forthlib@001A
		XDEF    forthlib@001B
		XDEF    forthlib@001C
		XDEF    forthlib@001D
		XDEF    forthlib@001E
		XDEF    forthlib@001F
		XDEF    forthlib@0020
		XDEF    forthlib@0021
		XDEF    forthlib@0022
		XDEF    forthlib@0023
		XDEF    forthlib@0024
		XDEF    forthlib@0025
		XDEF    forthlib@0026
		XDEF    forthlib@0027
		XDEF    forthlib@0028
/******************************************************************************/

/******************************************************************************/

/** cursor blink delay */
#define	BLINK_DELAY 4

/** Text output controll characters */
#define NUL 	0
#define INV	1/* invert text color */
#define UL	2/* toggle underlining */
#define BOLD	3/* toggle bold */
#define BEL 	7
#define BS	8
#define TAB	9
#define	LF	10
#define	FF	12
#define	CR	13

/** Tabulator length */
#define TAB_LEN 8
#define TAB_MASK %11111111-TAB_LEN+1
/*	Valid tabulator lengths are 2,4,8,16 */

/** underlining mask */
#define	UL_MASK %11111100

/** length of one symbol entry */
#define	sizeof_SYM_ENTRY (tios::SYM_ENTRY.hVal+2)

/** maximum Block Number */
#define BLOCK_NUM 9999
/******************************************************************************/

/******************************************************************************
 * Some tools that are often used:
 */

/** prepare for interrupt redirection */
PrepIntRedir:	move.w	$700, d0			// switch off ints
		trap	#1
		bclr	#2, $600001			// switch protected memory violation off
		rts
/** switch ints and protected memory violation on, after interrupt redirection is done */
EndIntRedir:	bset	#2, $600001
		trap	#1
		rts

/** execute tios::DEREF d0,a1  (function calls are smaller than this macro) */
DEREF_d0_a1:	tios::DEREF	d0,a1
		rts

/******************************************************************************
 * CURSOR/OUTPUT ROUTINES
 ******************************************************************************/

/*-----------------------------------------------------------------------------
 * Blink the cursor on the screen (also changes the Phase-flag)
 * Input: a0 = pointer to Cursor structure
 * Does NOT destroy any register!
 */
BlinkCursor:	movem.l	d0-d3/a0, -(a7)

		not.b	Cursor.Phase-Cursor(a0)		// change the phase 00->ff or ff->00

		movem.w	(a0), d0-d2			// d0 = X-coord; d1 = Y-coord; d2 = Height-1
		movea.l	Screen.Addr, a0			// a0 = *VideoRAM

		/******* calculate the offset and Mask of the cursor */
		addq.w	#7, d1				// d1 = Y-coord of bottom of cursor
		mulu	#30, d1				// d1 = line-offset of -"-
		move.w	d0, d3				// d3 = number of byte within line
		lsr.w	#3, d3				
		add.w	d3, d1				// d1 = byte offset within screen
		adda.w	d1, a0				// a0 = absolute byte offset of cursor
		andi.w	#$0007, d0			// d0 = pixel number within byte
		move.w	#%1111110000000000, d1		// d1 = cursor-mask
		lsr.w	d0, d1				// d1.b = cursor mask 2
		move.w	d1, d0				// d0.b = cursor mask 1
		lsr.w	#8, d0
		
		/******* invert the cursor */
\Loop:		eor.b	d0, (a0)+
		eor.b	d1, (a0)+
		lea	-32(a0), a0			// go to previous line
		dbra	d2, \Loop

 		movem.l	(a7)+, d0-d3/a0
		rts

/*-----------------------------------------------------------------------------
 * My cursor handler:
 */
CursorHandler:	move.l	a0, -(a7)
		lea	Cursor, a0			// a0 = pointer to Cursor-structure
		tst.w	Cursor.Hidden-Cursor(a0)	// Cursor.Hidden or .Off ? -> do nothing
		bne	\End
		subq.b	#1, Cursor.Count-Cursor(a0)	// decrease Cursor.Count
		bne	\End				//   cursor is only blinked, when counter reaches zero

		/****** blink the cursor */
		bsr	BlinkCursor			// blink the cursor
		move.b	#BLINK_DELAY, Cursor.Count-Cursor(a0)	// reset Cursor.Count
\End:		movea.l	(a7)+, a0
PtrOldAutoInt5:	jmp	$123456				// here the address of the old int is stored!

/*-----------------------------------------------------------------------------
 * My grayscale handler:
 */
GrayHandler:	move.l	a0, -(a7)
		move.w	d0, -(a7)
		lea	Screen.Phase, a0		// a0 = pointer to Screen-structure
		subq.b	#1, (a0)			// decrease phase
		bpl	\PhaseNonzero
		move.b	#11, (a0)			// if phase was zero: phase = 11
\PhaseNonzero:	
		move.w	Screen.IOAddr-Screen.Phase(a0), d0  // d0 = current video-RAM: plane1
		cmpi.b	#7, (a0)
		ble	\Plane1				    // if Phase > 7:
		move.w	Screen.IOAddr2-Screen.Phase(a0), d0 //   display plane 0 (else plane 1)
\Plane1:	move.w	d0, $600010			// adjust plane
	
		move.w  (a7)+, d0
		movea.l	(a7)+, a0
PtrOldAutoInt1:	jmp	$123456				// here the address of the old int is stored!

/*-----------------------------------------------------------------------------
 * Allocate and clear a screen
 * Input:  d0.w = handle or zero on error
 *         d1.l = pointer to screen (address is multiple of 8)
 *         d2.w = IO-Address (address / 8)
 */
AllocScreen:					// allocate screen
		pea	(3847).w			// shorter than move.l
		jsr	tios::HeapAlloc			
		addq.l	#4, a7
		tst.w	d0
		beq	\Error
						// get address
		move.w	d0, d1
		bsr	DEREF_d0_a1			// a1 = pointer to screen
		move.w	d1, d0
		move.l	a1, d2				// d2 = address / 8
		addq.l	#7, d2
		lsr.l	#3, d2
		move.l	d2, d1				// d1 = pointer to screen
		lsl.l	#3, d1
		
						// clear screen
		movea.l	d1, a0				// a0 = pointer to screen
		move.w	d3, -(a7)
		move.w	#30*128/4-1, d3			// d3 = counter
\Loop:		clr.l	(a0)+
		dbra	d3, \Loop
		move.w	(a7)+, d3		
		
\Error:		rts
		 
/******************************************************************************
 * CreateScreen
 *-----------------------------------------------------------------------------
 * Creates a new screen, the FORTH screen, and activates the cursor.
 * This routine has to be called before any FORTH output routines can 
 * be used.
 *-----------------------------------------------------------------------------
 * Input: none
 * Ouput:
 *	d0.l: 0 = OK; 1 = failure
 */
forthlib@0000:
CreateScreen:	/******* allocate&clear screen, initialize screen-record */
		bsr	AllocScreen			// allocate a new screen
		lea	Screen, a0			// a0 = pointer to screen data record
		move.w	d0, (a0)+			// set Screen.Handle
		beq	\Failure
		move.l	d1, (a0)+			// set Screen.Addr
		move.w	d2, (a0)+			// set Screen.IOAddr
		move.w	d2, $600010			// switch to that screen
		clr.w	(a0)+				// clear Screen.Handle2 (no grayscale)
		clr.l	(a0)+				//     and Screen.Addr2 (no grayscale)

		/******* initialize cursor-record */
		lea	Cursor, a0			// a0 points to cursor-data-record
		clr.l	(a0)+				//  .ScrX = .ScrY = 0
		move.w	#1, (a0)+			//  .Height = 1 (actual Height = 2 pixel)
		clr.l	(a0)+				//  .Hidden = false .Off = false
							//  .Phase = 0 (off) .TxtColor = 0 (B on W)
		clr.w	(a0)+				//  .ULflag = 0 (no underline) .BoldFlag = 0
		move.b	#1, (a0)+			//  .Count = 1

		/******* install my cusor hander */
		bsr	PrepIntRedir
		lea	PtrOldAutoInt5+2(PC), a1	// a1 points to field, where to store the old address
		move.l  $74, (a1)			// save the address of AutoInt5 int (a1)
                lea	CursorHandler(PC), a1		// set AutoInt5 to my Cursor-Handler
		move.l	a1, $74
		bsr	EndIntRedir

		moveq	#0, d0				// d0 = 0: all ok
		rts
\Failure:	moveq	#1, d0				// d0 = 1: failure
		rts

/******************************************************************************
 * GraphicOn
 *-----------------------------------------------------------------------------
 * Switch the standard text screen, created by `CreateScreen' to a 4 grayscale
 * graphic screen. The cursor will be switched off, text that is outputted by
 * `Emit' will be Dark Grey. If `GraphicOn' had already been called before, 
 * this operation will have no effect.
 *-----------------------------------------------------------------------------
 * Input:  none
 * Output: d0 = zero (success) or d0 = nonzero (failure)
 */
forthlib@0001:
GraphicOn:	move.l	a3, -(a7)
		lea	Screen.Handle2, a3		// a3 = pointer to screen data record
		tst.w	(a3)				// grayscale allready active ?
		bne	\End				//   -> do nothing
		
		bsr	AllocScreen			// allocate a new screen
		move.w	d0, (a3)+			// set Screen.Handle2
		beq	\Failure
		move.l	d1, (a3)+			// set Screen.Addr2
		move.w	d2, (a3)+			// set Screen.IOAddr2
		clr.b	(a3)+				// set Screen.Phase = 0
		bsr	HideCursor			// hide cursor 
		st	Cursor.Off			// and switch it off
		
		/** Install grayscale handler into Auto Int 1 */
		bsr	PrepIntRedir
		lea	PtrOldAutoInt1+2(PC), a1	// a1 points to field, where to store the old address
		move.l  $64, (a1)			// save the address of AutoInt1 int (a1)
                lea	GrayHandler(PC), a1		// set AutoInt1 to my grayscale handler
		move.l	a1, $64
		bsr	EndIntRedir

		clr.w	d0				// Success: d0 = zero
\End:		movea.l	(a7)+, a3
		rts
\Failure:	moveq	#1, d0				// Failure: d0 = nonzero
		bra	\End

/******************************************************************************
 * GraphicOff
 *-----------------------------------------------------------------------------
 * Switch back to the text screen, and enable the cursor. If there was no
 * graphic screen active this routine will have no effect.
 *-----------------------------------------------------------------------------
 * Input:  none
 * Output: none
 */
forthlib@0002:
GraphicOff:	lea	Screen.Handle2, a0		
		move.w	(a0), d0			// d0 = handle of 2nd plane
		beq	\NoGrayscale		  	// if zero: grayscale is already off
		clr.w	(a0)+				// set handle to zero
		clr.l	(a0)+				// also clear address
		bsr	HeapFree			// delete 2nd grayplane
		bsr	PrepIntRedir			// restore the original AutoInt1
		move.l	PtrOldAutoInt1+2(PC), $64
		bsr	EndIntRedir
		move.w	Screen.IOAddr, $600010 		// switch to text screen
		
		sf	Cursor.Off			// enable cursor
		bsr	ShowCursor			// and show it
\NoGrayscale:	rts
		
/******************************************************************************
 * DestroyScreen
 *-----------------------------------------------------------------------------
 * Destroys the FORTH screen, switches off cursor etc. This routine may
 * be called if the graphic screen is active.
 *-----------------------------------------------------------------------------
 * Input: none
 * Ouput:
 *	old screen is restored
 */
forthlib@0003:
DestroyScreen:	/******* switch off cursor */
		bsr	PrepIntRedir
		move.l	PtrOldAutoInt5+2(PC), $74	
		bsr	EndIntRedir

		/******* restore screen */
		move.l	#tios::main_lcd, d0		// switch back to the standart TIOS-screen
		lsr.l	#3, d0
		move.w	d0, $600010

		move.w	Screen.Handle, d0		// free the video RAM
		bsr	HeapFree

		bra	GraphicOff			// switch graphic screen off

/******************************************************************************
 * HideCursor
 *-----------------------------------------------------------------------------
 * Hides the cursor, until ShowCursor is called. If HideCursor is called
 * many times, the cursor will stay invisible, until ShowCursor was called the
 * same number of times. I fthe cursor is currently off, do nothing.
 *-----------------------------------------------------------------------------
 * Input: none
 * Ouput:
 *	cursor is hidden
 * Destroys only a0
 */
forthlib@0004:
HideCursor:	lea	Cursor, a0
		tst.b	Cursor.Off-Cursor(a0)
		bne	\CursorOff
		addq.b	#1, Cursor.Hidden-Cursor(a0)	// increase hidden-flag
		tst.b	Cursor.Phase-Cursor(a0)
		beq	\CursorOff			// if the cursor is currently on
		bsr	BlinkCursor			//  -> switch it off
\CursorOff:	rts


/******************************************************************************
 * ShowCursor
 *-----------------------------------------------------------------------------
 * Shows the cursor, after it was hidden by HideCursor. If the cursor is 
 * already visible or is switched off, do nothing.
 *-----------------------------------------------------------------------------
 * Input: none
 * Ouput:
 *	cursor is shown 
 * Destroys only a0
 */
forthlib@0005:
ShowCursor:	lea	Cursor, a0		
		tst.b	Cursor.Hidden-Cursor(a0)	// if cursor is allready there:
		beq	\End				//   --> do nothing
		tst.b	Cursor.Off-Cursor(a0)		// if it's off:
		bne	\End				//   --> do nothing
		subq.b	#1, Cursor.Hidden-Cursor(a0)	// decrease hidden-flag
		bne	\End				// if it is still not zero: cursor stays hidden
		bsr	BlinkCursor			// show cursor
		move.b	#BLINK_DELAY, Cursor.Count-Cursor(a0)
\End:		rts


/******************************************************************************
 * SetCursor
 *-----------------------------------------------------------------------------
 * Sets the cursor to a specified text-screen position
 *-----------------------------------------------------------------------------
 * Input:
 * 	d0.w = x-character-coord
 *      d1.w = y-character-coord
 *      (coords are checked for screen boundaries)
 * Ouput: none
 * be careful: d0/d1 are destroyed
 */
forthlib@0006:
SetCursor:	tst.w	d0				// x < 0 ?
		bpl	\XlowOk
		clr.w	d0				//   -> x = 0
\XlowOk:	tst.w	d1				// y < 0 ?
		bpl	\YlowOk
		clr.w	d1				//   -> y = 0
\YlowOk:	cmpi.w	#40, d0				// x >= 40 ?
		blt	\XuppOk	
		moveq	#39, d0				//   -> x = 39
\XuppOk:	cmpi.w	#16, d1				// y >= 16
		blt	\YuppOk
		moveq	#15, d1				//   -> y = 15
\YuppOk:
		bsr	HideCursor
		lea	Cursor, a0
		mulu	#6, d0				// d0 *= 6 (pixel-position)
		lsl.w	#3, d1				// d1 *= 8 (pixel-position)
		move.w	d0, (a0)+			// store d0/d1 into Cursor.Scrx and Cursor.ScrY
		move.w	d1, (a0)+
		bsr	ShowCursor
		rts

/******************************************************************************
 * GetCursor
 *-----------------------------------------------------------------------------
 * Returns the text position of the cursor
 *-----------------------------------------------------------------------------
 * Input: none
 * Ouput:
 * 	d0.w = x-character-coord
 *      d1.w = y-character-coord
 * Only a0 is destroyed
 */
forthlib@0007:
GetCursor:	lea	Cursor.ScrX, a0		
		clr.l	d0				// divu processes 32bits!!! (this was a terrible bug)
		move.w	(a0)+, d0			// load Screen-X-coord
		divu	#6, d0				// calculate character-X-coord
		move.w	(a0)+, d1			// load Screen-Y-coord
		lsr.w	#3, d1				// calculate character-Y-coord
		rts

/******************************************************************************
 * CursorHeight
 *-----------------------------------------------------------------------------
 * Sets the cursor's height
 *-----------------------------------------------------------------------------
 * Input: 
 *	d0.b = Cursor height, must be in the range 1..8
 * Ouput:
 * 	cursor is resized
 * be careful: d0 is destroyed!
 */
forthlib@0008:
CursorHeight:	tst.b	d0				// height <= 0 ?
		bgt	\LowOk
		moveq	#1, d0				//  -> height = 1
\LowOk:		cmpi.w	#8, d0				// height > 8 ?
		ble	\UppOk
		moveq	#8, d0				//  -> height = 8
\UppOk:
 		bsr	HideCursor
		subq.w	#1, d0				// store height-1 in Cursor.Height
		move.w	d0, Cursor.Height
		bra	ShowCursor

/******************************************************************************
 * Scroll1Up Scroll1Down Scroll1Left Scroll1Right
 *-----------------------------------------------------------------------------
 * Scroll the screen by one row/column
 *-----------------------------------------------------------------------------
 * Input: none
 * Output: 
 *   	screen is scrolled
 */
forthlib@0009:
Scroll1Up:	bsr	HideCursor
		move.l	Screen.Addr, a0			// load screen address
		move.w	#900-1, d0			// move 900 longwords
\scrollLoop:	move.l	30*8(a0), (a0)+
		dbra	d0, \scrollLoop
		move.w	#8*30-1, d0			// clear 8 lines with 30 bytes
		move.b	Cursor.TxtColor, d1		// get color, to use for clearing
\clearLoop:	move.b	d1, (a0)+
		dbra	d0, \clearLoop
		bra	ShowCursor

forthlib@000A:
Scroll1Down:	bsr	HideCursor
		move.l	Screen.Addr, a0			// load screen address
		lea	3840(a0), a0			// end of screen
		move.w	#900-1, d0			// move 900 longwords
\scrollLoop:	move.l	-30*8-4(a0), -(a0)
		dbra	d0, \scrollLoop
		move.w	#8*30-1, d0			// clear 8 lines with 30 bytes
		move.b	Cursor.TxtColor, d1		// get color, to use for clearing
\clearLoop:	move.b	d1, -(a0)
		dbra	d0, \clearLoop
		bra	ShowCursor
	
forthlib@000B:
Scroll1Left:	bsr	HideCursor
		move.l	Screen.Addr, a0			// load screen address
		moveq	#128-1, d0			// process 128 rows
\YLoop:		moveq	#15-1, d1			//   process 15 words = 240 columns
\__XLoop:	move.l	(a0), d2			//     load 32 pixels
		lsl.l	#6, d2				//     shift them 6 pixels to the left
		swap	d2				//     write the right 16 pixels back
		move.w	d2, (a0)+
		dbra	d1, \__XLoop			//   ^^

		andi.b	#%11000000, -(a0)		//   clear the last 6 pixels, that are from the next row
		move.b	Cursor.TxtColor, d2		//   set the new column to the right color
		lsr.b	#2, d2
		eor.b	d2, (a0)+
		
		dbra	d0, \YLoop			// ^^
		bra	ShowCursor

forthlib@000C:
Scroll1Right:	bsr	HideCursor
		move.l	Screen.Addr, a0			// load screen address
		lea	3840(a0), a0			// end of screen +2
		moveq	#128-1, d0			// process 128 rows
\YLoop:		moveq	#15-1, d1			//   process 16 words = 240 columns
\__XLoop:	move.l	-4(a0), d2			//     load 32 pixels
		lsr.l	#6, d2				//     shift them 6 pixels to the right
		move.w	d2, -(a0)			//     write the right 16 pixels back
		dbra	d1, \__XLoop			//   ^^
		andi.b	#%00000011, (a0)		//   clear the 6 pixels, that are from the previous row
		move.b	Cursor.TxtColor, d2		//   set the new column to the right color
		lsl.b	#2, d2
		eor.b	d2, (a0)
		dbra	d0, \YLoop			// ^^
		bra	ShowCursor

/******************************************************************************
 * Emit
 *-----------------------------------------------------------------------------
 * Print a character onto the screen, at the current cursor position and with
 * the current color. Adjust the cursor one column to the right afterwards.
 * If the cursor has reached the end of line; put it to the next line.
 * If the cursor has reached the end of screen, scroll screen one row upwards.
 *-----------------------------------------------------------------------------
 * Input: 
 *	d0.w = character (high-order bits are ignored)
 *             Controllcharacters allowed!
 * Ouput:
 * 	character is drawn
 */
forthlib@000D:
Emit:		move.w	d3, -(a7)
		bsr	HideCursor

		clr.w	d3
		move.b	d0, d3				// d3 = character
		bsr	GetCursor			// d0|d1 : cursor position
		move.l	Screen.Addr, a0			// a0 = pointer to screen
		lea	Cursor, a1			// a1 = pointer to cursor-record
		move.b	Cursor.TxtColor-Cursor(a1), d2	// d2 = color (0: b on w, $FF: w on b)

		/******* Check for control characters */
		cmpi.b	#31, d3
		bhi	\NotCtrlChar

		cmpi.b	#INV, d3			// INV (invert textcolor)
		bne	\NotINV
		not.b	Cursor.TxtColor-Cursor(a1)
		bra	\End

\NotINV:	cmpi.b	#UL, d3				// UL (toggle underlining)
		bne	\NotUL
		not.b	Cursor.ULflag-Cursor(a1)
		bra	\End

\NotUL:		cmpi.b	#BOLD, d3			// BOLD (toggle bold)
		bne	\NotBOLD
		not.b	Cursor.BoldFlag-Cursor(a1)
		bra	\End

\NotBOLD:	cmpi.b	#CR, d3				// CR (go to begin of current line)
		beq	\DoCR

		cmpi.b	#LF, d3				// LF (go to begin of next line)
		beq	\DoLF

		cmpi.b	#BS, d3				// BS (go one character backwards)
		bne	\NotBS
		subq.w	#1, d0
		bge	\BSOk
		moveq	#39, d0
		subq.w	#1, d1
		bge	\BSOk
		bsr	Scroll1Up
\BSOk:		bra	\AdjustCursor

\NotBS:		cmpi.b	#TAB, d3			// TAB
		bne	\NotTAB
		addq.w	#TAB_LEN, d0
		andi.b	#TAB_MASK, d0
		bra	\CheckCursor

\NotTAB:	cmpi.b	#BEL, d3			// BEL (flash screen)
		bne	\NotBEL
		move.w	#960-1, d0
		move.w	d0, d1
\BELloop1:	not.l	(a0)+				//     invert 
		dbra	d1, \BELloop1
\BELloop2:	dbra	d1, \BELloop2			//     wait
\BELloop3:	not.l	-(a0)				//     invert back
		dbra	d0, \BELloop3
		bra	\End

\NotBEL:	cmpi.b	#FF, d3				// FF (clear screen)
		bne	\NotFF
		ext.w	d2				//     get background-longword
		ext.l	d2
		move.w	#960-1, d0			//     clear the screen
\FFLoop:	move.l	d2, (a0)+
		dbra	d0, \FFLoop
		clr.w	d0				//     adjust the cursor to (0|0)
		clr.w	d1
		bra	\AdjustCursor
\NotFF:
\NotCtrlChar:
		/******* Calculate addresses & co */
		lea 	tios::SF_font+$0600, a1		// a1 = pointer to Font table
		lsl.w	#3, d3				// a1 = pointer to letter
		adda.w	d3, a1

		mulu	#30*8, d1			// a0 = pointer to character within VideoRAM
		adda.w	d1, a0
		mulu	#6, d0
		move.w	d0, d1				
		lsr.w	#3, d1
		adda.w	d1, a0
		andi.b	#%00000111, d0			// d0 = pixel offset within byte

		lsl.b	#2, d2				// d2 = invert mask for changing the color of a char.

		/******* Draw character */
		moveq	#8-1, d1			// output the 8 rows of the character
\Loop:		move.w	#%1111111100000011, d3		//   clear the area, where the character is drawn
		ror.w	d0, d3
		and.b	d3, (a0)
		lsr.w	#8, d3
		and.b	d3, 1(a0)

		clr.w	d3	
		move.b	(a1)+, d3			//   d3.w = current row of character
		tst.w	d1				//   Underline character
		bne	\__NoUL				//    if last row and
		tst.b	Cursor.ULflag			//    underline flat is set
		beq	\__NoUL
		or.b	#UL_MASK, d3			//   add an underline to the chr row
\__NoUL:	tst.b	Cursor.BoldFlag			//   Print bold character
		beq	\__NotBold
                move.w	d3, -(a7)			
		lsl.b	#1, d3
		or.w	(a7)+, d3
\__NotBold:	eor.b	d2, d3				//   change char to current color		
		ror.w	d0, d3				//   shift it to the right pixel-offset
		or.b	d3, (a0)+			//   set the black pixels
		lsr.w	#8, d3
		or.b	d3, (a0)
		
		adda.w	#29, a0				//   go to next screen-row
		dbra	d1, \Loop			// ^^ 

		/******* Adjust the cursor */
		bsr	GetCursor			// set cursor to next character
		addq.w	#1, d0
\CheckCursor:	cmpi.w	#40, d0				// Column >= 40 ?
		blt	\AdjustCursor
\DoLF:		addq.w	#1, d1				//   -> to next row
\DoCR:		clr.w	d0				//   -> to first column
		cmpi.w	#16, d1				//   -> Row >= 16
		blt	\AdjustCursor
		bsr	Scroll1Up			//      -> scroll one row upwards
		moveq	#15, d1				//      -> set up registers (that were just destroyed)
		clr.w	d0
\AdjustCursor:	bsr	SetCursor

\End:		move.w	(a7)+, d3
		bra	ShowCursor


/******************************************************************************
 * FILE ACCESS ROUTINES
 ******************************************************************************/

/*-----------------------------------------------------------------------------
 * IsValidName
 *-----------------------------------------------------------------------------
 * Returns, whether a variable/folder name is valid. Valid names may begin on
 * a letter and consist of 'a'-'z', '0'-'9' and '_'. The maximum name length is
 * 8. Note that only lowercase names are valid.
 *-----------------------------------------------------------------------------
 * Input:
 *	a0 = pointer to zero terminated name
 * Output:
 *	d0.w = 1 (valid) or 0 (invalid)
 * 	Z-flag represents state of d0
 * Register d0-d1/a0 destroyed
 */
IsValidName:	moveq	#8, d0				// d0 = length counter

\Loop:		move.b	(a0)+, d1			// d1 = character
		beq	\Valid
		subq.w	#1, d0				// decrese and check length counter
		bmi	\Invalid
		cmpi.b	#'z', d1			// check 'a'-'z'
		bhi	\Invalid
		cmpi.b	#'a', d1
		bcc	\Loop
		cmpi.w	#7, d0				// if this was the first character, that 
		beq	\Invalid			// hasn't been 'a'-'z' the name is invalid

		cmpi.b	#'_', d1			// check '_'
		beq	\Loop
		cmpi.b	#'9', d1			// check '0'-'9'
		bhi	\Invalid
		cmpi.b	#'0', d1
		bcc	\Loop				

\Invalid:	moveq	#0, d0				// code for invalid
		rts
\Valid:		moveq	#1, d0				// code for valid
		rts


/*-----------------------------------------------------------------------------
 * FindSymEntry
 *-----------------------------------------------------------------------------
 * Search for the folder/file given by a0 in the symbol table, specified by d0
 * and return its handle.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to file/folder name
 *	d0   = handle of file / folder table
 * Ouput:
 *      a0.l = pointer to symbol entry or zero on error
 *      d0.w = handle or zero on error (not found or invalid name)
 *      Z-flag represents the state of d0
 */
FindSymEntry:		// SYM_ENTRY *FindSymEntry(WORD symlist, BYTE *name)
		pea	(a0)
		move.w	d0, -(a7)
		bsr	IsValidName			// check the file/folder name (a0)
		beq	\Error
		
		jsr	tios::FindSymEntry
		move.l	a0, d0				
		beq	\Error
		move.w	tios::SYM_ENTRY.hVal(a0), d0

\Error:		adda.w	#6, a7				// adda doesn't modify the flags!
		rts
/*-----------------------------------------------------------------------------
 * FindSymEntry
 *-----------------------------------------------------------------------------
 * Search for the folder given by a0 in the folder table its handle or zero
 * on error.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to folder name
 *	d0   = handle of folder table
 * Ouput:
 *      a0.l = pointer to symbol entry or zero on error
 *      d0.w = handle or zero on error (not found or invalid name)
 *      Z-flag represents the state of d0
 */
FindFolder:	move.w	FOLDER_LIST_HANDLE, d0		// get handle of folder
		bra	FindSymEntry			// rts in FindSymEntry

/*-----------------------------------------------------------------------------
 * getSymEntry
 *-----------------------------------------------------------------------------
 * Find an entry in the file/folder table. If no matching entry can be found,
 * a new entry is created. That new entry's handle will be zero.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of entry
 *      d0.w = handle of file/folder table
 * Ouput:
 * 	a0.l = pointer to entry
 *      d0.w = handle of entry or -1 on error (not enough memory or invalid name)
 *             if the file was newly created it will be zero
 *      no registers are destroyed!
 */
getSymEntry:	movem.l	d1-d7/a1-a6, -(a7)
		
\Aname		EQUR	a2				// pointer to name of entry
		movea.l	a0, \Aname
\Dhdl		EQUR	d3				// handle of symbol table
		move.w	d0, \Dhdl	
\Asym		EQUR	a3				// pointer to symbol table
		bsr	DEREF_d0_a1
		movea.l	a1, \Asym
\Awsym		EQUR	a4				// working pointer into sym table
		lea	2(\Asym), \Awsym
\Delms		EQUR	d4				// number of elements in table
		move.w	(\Awsym)+, \Delms		

		bsr	IsValidName			// check the varname (a0)
		beq	\Error

\Csym		EQUR	d5				// counter for searching the table
		move.w	\Delms, \Csym
		subq.w	#1, \Csym			// if table is empty: name not found!
		bmi	\NotFound

	/******** Search for the entry */
\SymLoop:	// ________________________________
\Acmp1		EQUR	a5				// comparison pointer
\Acmp2		EQUR	a6
		movea.l	\Awsym, \Acmp1
		movea.l	\Aname, \Acmp2
		
\__CmpLoop:	//     __________________________	// compare current entry with name
		cmp.b	(\Acmp2)+, (\Acmp1)+
		bhi	\NotFound			// greater
		bcs	\__NextEntry			// less than
		tst.b	(\Acmp2)
		bne	\__CmpLoop
		//     ^^^^^^^^^^^^^^^^^^^^^^^^^^
 		
	/******** we found the entry */
		movea.l	\Awsym, a0			// a0 := entry
		move.w	tios::SYM_ENTRY.hVal(a0), d0	// d0 := handle
		bra	 \End

\__NextEntry:	lea	sizeof_SYM_ENTRY(\Awsym), \Awsym // goto next symbol entry
		dbra	\Csym, \SymLoop
		// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

	/******** the entry doesn't exist, so we have to create a new one. */
\NotFound:	/* Csym is the number of entries to be moved-1!
		 * Awsym is the pointer to the place where the new entry has to be created. 
		 */
		cmp.w	(\Asym), \Delms			// is there enough space in the table?
		blt	\DontRealloc

		/****** resize the symbol table ********/
		move.w	\Delms, d0			// d0 = new size of symbol table
		addq.w	#8, d0
		mulu	#sizeof_SYM_ENTRY, d0
		addq.l	#4, d0

			// WORD HeapRealloc(WORD handle, LONG newsize)
		move.l	d0, -(a7)			
		move.w	\Dhdl, -(a7)		
		jsr	tios::HeapRealloc
		addq.l	#6, a7
		
		tst.w	d0
		beq	\Error
		bsr	DEREF_d0_a1
		suba.l	\Asym, \Awsym			// Awsym is now a relative pointer
		movea.l	a1, \Asym			// Asym points to resized table
		adda.l	\Asym, \Awsym			// Awsym also points into resized table
		add.w	#8, (\Asym)			// increase max-entries field of symtable
\DontRealloc:
		/****** move entries *******************/
		tst.w	\Csym				// Csym is #entries to be moved-1!
		bmi	\DontMove			// (new entry at end of list!)
\Aend		EQUR	a5				// pointer to end of table
		move.w	\Csym, d0			// (move from end)
		mulu	#sizeof_SYM_ENTRY, d0
		lea	sizeof_SYM_ENTRY(\Awsym,d0), \Aend
\Cmove		EQUR	d7				// number of words to move - 1 (counter)
		move.w	\Csym, \Cmove
		mulu	#sizeof_SYM_ENTRY/2, \Cmove
		addq.w	#sizeof_SYM_ENTRY/2-1, \Cmove
		
		// ________________________________	// do the move
\MoveLoop:	move.w	-(\Aend), sizeof_SYM_ENTRY(\Aend)
		dbra	\Cmove, \MoveLoop
		// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
\DontMove:	
		/******* create the new entry **********/
		addq.w	#1, 2(\Asym)			// increase the number-of-entries-field
		movea.l	\Awsym, a0			// a0 := entry
		clr.w	d0				// d0 := handle
		moveq	#7, d7				// set the name field
\SetNameLoop:	move.b	(\Aname)+, (\Awsym)+
		dbra	d7, \SetNameLoop
		clr.l	(\Awsym)			// set flags and handle to zero

\End:		movem.l	(a7)+, d1-d7/a1-a6
		rts
		
\Error:		moveq	#-1, d0				// d0 = errorcode
		bra	\End

/*-----------------------------------------------------------------------------
 * DeleteSymEntry
 *-----------------------------------------------------------------------------
 * Search for the folder/file given by a0 in the symbol table, specified by d0,
 * remove it from the table and return its handle.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to file/folder name
 *	d0   = handle of file/folder table
 * Ouput:
 *      d0.w = handle of removed entry or zero on error (file not found or invalid name)
 * no registers are destroyed!
 */
DeleteSymEntry:	movem.l	d1-d7/a0-a6, -(a7)
		move.w	d0, d1				// d1 = handle of table
		bsr	DEREF_d0_a1			// a1 = pointer to folder table
		addq.l	#2, a1
		movea.l	a1, a2				// a2 = pointer to number of files
		move.w	(a1)+, d3			// d3 = number of files in folder
		mulu	#sizeof_SYM_ENTRY, d3
		lea	-sizeof_SYM_ENTRY(a1, d3.w), a3	// a3 = pointer to last file in folder
		
		move.w	d1, d0				// search for the file in the folder	
		bsr	FindSymEntry			// d0 = handle of file; a0 = sym entry
		beq	\End				// Z-flag = 1: file not found or invalid name

		/****** remove the file from the list **/
		subq.w	#1, (a2)			// decrease the file count in the table
\RemoveLoop:	cmpa.l	a3, a0				// a3 = pointer to begin of last file
		beq	\End
		move.l	sizeof_SYM_ENTRY(a0), (a0)+
		bra	\RemoveLoop
		
\End:		movem.l	(a7)+, d1-d7/a0-a6
		rts

/*-----------------------------------------------------------------------------
 * SplitPath
 *-----------------------------------------------------------------------------
 * Split a path "folder\varname" into "folder" and "varname" and check whether
 * it is correct (uses the right characters, has a correct length etc.)
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to path
 * Ouput:
 *      d0.w = NONZERO on success, else zero (too long foldername)
 *             NOTE: this differs from other routines' return flags!
 *      Z-flag represents the state of d0
 * 	a2.l = pointer to varname
 *      a3.l = pointer to folder
 */
SplitPath:	lea	TempBuffer, a3			// a3 = copy of folder name
		movea.l	a3, a1				// a1 = copy pointer 
			
		/** search for '\' */
		moveq	#8, d2				// d2 = folder name length counter
\SearchLoop:	move.b	(a0)+, (a1)			// copy character
		beq	\Invalid
		cmpi.b	#'\', (a1)
		beq	\Backslash
		addq.l	#1, a1
		dbra	d2, \SearchLoop
\Invalid:	moveq	#0, d0				// d0 = error!
		rts
		
\Backslash:	clr.b	(a1)				// terminate folder name ('\' --> '\0')
		movea.l	a0, a2				// set name-pointer after '\'

		moveq	#1, d0				// d0 = ok
		rts	

/*-----------------------------------------------------------------------------
 * EnterFolder
 *-----------------------------------------------------------------------------
 * Get the handle of a folder, given in the form "folder\varname" and return
 * the name component to be further processed.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to path
 * Ouput:
 *      d0.w = handle or zero on error
 *      Z-flag represents the state of d0
 *	a0.l = pointer to varname-component of folder
 * d1-d2/a1-a3 destroyed!
 */
EnterFolder:	bsr	SplitPath			// (a3) = folder; (a2) = varname
		beq	\Error				// Z-flag = 1: invalid path

		movea.l	a3, a0
		bsr	FindFolder			// d0 = handle of folder
		beq	\Error				// Z-flag = 1: invalid folder name

		movea.l	a2, a0				// a0 = name component
\Error:		rts

/*-----------------------------------------------------------------------------
 * HeapFree
 *-----------------------------------------------------------------------------
 * Free the handle, specified by d0
 *-----------------------------------------------------------------------------
 * Input: 
 *	d0 = handle
 * Ouput:
 *	handle is freed.
 */
HeapFree:	move.w	d0, -(a7)
		jsr	tios::HeapFree
		addq.l	#2, a7
		rts

/******************************************************************************
 * CreateFile
 *-----------------------------------------------------------------------------
 * Create a new file, the name given by a0 in the format "folder\varname".
 * If the file exists, it will be deleted and recreated.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of file
 *	d0.w = handle to associate with it
 * Ouput:
 * 	d0.w = zero on success, else nonzero
 *      a0.l = symbol entry of the newly created file
 *      Z-flag represents state of d0
 * No registers modified!
 */
forthlib@000E:
CreateFile:	movem.l	d1-d7/a1-a6, -(a7)
		move.w	d0, d3				// d3 = handle of new file

		bsr	EnterFolder			// d0 = handle; a2 = varname
		beq	\Error				// d0=0 and Z-flag set: error 
		bsr	getSymEntry			// find/create file
		tst.w	d0				// d0 = handle; is negative on error
		bmi	\Error
		move.w	d3, tios::SYM_ENTRY.hVal(a0)	// replace handle of file by new handle
		clr.w	tios::SYM_ENTRY.flags(a0)	// clear flags
						// delete the original file's handle
		tst.w	d0				// if d0=0: file has no handle (new file)
		beq	\DontFree
		move.l	a0, -(a7)
		bsr	HeapFree			// free the handle d0
		movea.l	(a7)+, a0
\DontFree:		
		clr.w	d0				// d0 = success!
\End:		movem.l	(a7)+, d1-d7/a1-a6
		rts
		
\Error:		moveq	#1, d0				// d0 = errorcode
		bra	\End

/******************************************************************************
 * OpenFile
 *-----------------------------------------------------------------------------
 * Get a file's handle, the name given by a0 in the format "folder\varname".
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of file
 * Ouput:
 * 	d0.w = handle or zero on error
 *	a0.l = pointer to symbol entry of file
 * 	Z-flag represents state of d0
 * No registers modified!
 */
forthlib@000F:
OpenFile:	movem.l	d1-d7/a1-a6, -(a7)
		bsr	EnterFolder			// d0 = handle; a2 = varname
		beq	\Error				// d0=0 and Z-flag set: error 
		bsr	FindSymEntry			// d0 = handle of file
\Error:		movem.l	(a7)+, d1-d7/a1-a6
		rts

/******************************************************************************
 * DeleteFile
 *-----------------------------------------------------------------------------
 * Delete a file, its name given by a0 in the form "folder\main".
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of file
 * Ouput:
 * 	d0.w = zero on success, nonzero else
 * No registers modified!
 */
forthlib@0010:
DeleteFile:	movem.l	d1-d7/a0-a6, -(a7)
		bsr	EnterFolder			// d0 = handle; a0 = varname
		beq	\Error				// d0=0 and Z-flag set: error 
		bsr	DeleteSymEntry			// delete file's entry (a0) in table d0
		tst.w	d0				// d0 = handle of removed entry
		beq	\Error
		bsr	HeapFree			// free the handle d0
		moveq	#0, d0				// d0 = success
\End:		movem.l	(a7)+, d1-d7/a0-a6
		rts
\Error:		moveq	#1, d0				// d0 = errorcode
		bra	\End

/******************************************************************************
 * MoveFile
 *-----------------------------------------------------------------------------
 * Move one file, its name given by a0 in the form "folder\varname", to the 
 * file, whose name is given by a1 in the same form. If the destination file
 * allready exists, it will be overwritten.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to source-name ("folder1\varname1")
 *      a1.l = pointer to destination-name ("folder2\varname2)
 * Ouput:
 * 	d0.w = zero on success, else nonzero (file not found, invalid name or
 *             not enough memory)
 *      No registers modified!
 */
forthlib@0011:
MoveFile:	movem.l	d1-d6/a0-a6, -(a7)
		movea.l	a0, a3				// save source-name
		bsr	OpenFile
		move.w	d0, d1				// d1 = handle of src file
		beq	\Error
		movea.l	a1, a0				// a0 = destination name
		bsr	OpenFile			// d0 = handle of dst file (if exists)
		cmp.w	d0, d1				// source handle == dest handle ? 
		beq	\Error

		movea.l	a1, a0				// create destination file with src hdl
		move.w	d1, d0
		bsr	CreateFile
		bne	\Error
		
		movea.l	a3, a0				// a0 = source-name
		bsr	EnterFolder			// d0 = folder-hdl, a0 = varname
		bsr	DeleteSymEntry			// delete the source, don't free its hdl

		moveq	#0, d0				// d0 = success
\End:		movem.l	(a7)+, d1-d6/a0-a6
		rts
\Error:		moveq	#1, d0				// d0 = error
		bra	\End

/******************************************************************************
 * CreateFolder
 *-----------------------------------------------------------------------------
 * Create a new folder, its name given by a0.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of folder
 * Ouput:
 * 	d0.w = zero on success, else nonzero (folder allready exists, invalid
 *             name or not enough memory)
 */
forthlib@0012:
CreateFolder:	move.l	a0, -(a7)			// save a0

		/******* create folder table ***********/
		pea	(8*sizeof_SYM_ENTRY+4).w	// shorter than move.l
		jsr	tios::HeapAlloc
		addq.l	#4, a7
		tst.w	d0
		beq	\Error
		
		move.w	d0, d1				// d1 = handle (d0 is destroyed)
		bsr	DEREF_d0_a1
		move.l	#$00080000, (a1)		// initialize folder table to be empty
		
		/******* create a new entry for that folder */
		movea.l	(a7)+, a0			// a0 = folder name
		move.w	FOLDER_LIST_HANDLE, d0		// d0 = handle of folder table
		bsr	getSymEntry			// a0 = pointer to new entry
		tst.w	d0				
		bne	\SymEntryError			// d0 != 0: folder allready exists ...
		bset.b	#7, tios::SYM_ENTRY.flags+1(a0)	// mark it as a folder
		move.w	d1, tios::SYM_ENTRY.hVal(a0)	// associate entry with folder handle
		rts					// d0 = 0 -- no error
		
\SymEntryError:	move.w	d1, d0				// free the handle
		bsr	HeapFree
\Error:		moveq	#1, d0				// d0 = 1 -- error
		rts
		
/******************************************************************************
 * DeleteFolder
 *-----------------------------------------------------------------------------
 * Delete a folder, its name given by a0.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to name of folder
 * Ouput:
 * 	d0.w = zero on success, else nonzero (folder not found, invalid name or
 *             folder not empty)
 *      Z-flag represents state of d0
 */
forthlib@0013:
DeleteFolder:	move.l	a0, -(a7)			// save a0
		bsr	FindFolder			// find folder
		beq	\Error
		bsr	DEREF_d0_a1			// a1 = pointer to folder contents
		
		tst.w	2(a1)				// check whether folder is empty
		bne	\Error
		
		movea.l	(a7)+, a0			// restore a0 (name of folder)
		move.w	FOLDER_LIST_HANDLE, d0		
		bsr	DeleteSymEntry			// delete the folder
		rts

\Error:		addq.l	#4, a7				// remove saved a0
		moveq	#1, d0
		rts

/******************************************************************************
 * RenameFolder
 *-----------------------------------------------------------------------------
 * Rename a folder. A folder with the destination name mustn't exist.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to source name
 *	a1.l = ponter to destination name
 * Ouput:
 * 	d0.w = zero on success, else nonzero (folder not found, invalid name or
 *             destination folder already exists)
 *	Z-flag represents state of d0
 */
forthlib@0014:
RenameFolder:	movem.l	d1-d7/a0-a6, -(a7)
		move.w	FOLDER_LIST_HANDLE, d3		// d3 = handle of folder list
		movea.l	a0, a3				// a3 = source name
		movea.l	a1, a4				// a4 = destination name
						// check, whether src folder exists
		move.w	d3, d0				// d0 = handle of folder list
		bsr	FindSymEntry			
		beq	\Error
						// create dest folder
		move.w	d3, d0				// d0 = handle of folder list	
		movea.l	a4, a0				// a0 = dest name
		bsr	getSymEntry			// create the destination folder
		tst.w	d0				// d0 = handle of existing entry
		bne	\Error				// if it already exists: error
		movea.l	a0, a5				// a5 = sym entry of new folder
						// delete the source folder
		move.w	d3, d0				// d0 = handle of folder list
		movea.l	a3, a0				// a0 = src name
		bsr	DeleteSymEntry			// delete source folder
						// assign file list to new folder
		move.w	d0, tios::SYM_ENTRY.hVal(a5)
		bset.b	#7, tios::SYM_ENTRY.flags+1(a5)	// make sym entry a folder

		clr.w	d0				// d0 = success
\End:		movem.l	(a7)+, d1-d7/a0-a6
		rts
\Error:		moveq	#1, d0				// d0 = errorcode
		bra	\End

/******************************************************************************
 * BLOCK READ/WRITE ROUTINES
 ******************************************************************************/

/*-----------------------------------------------------------------------------
 * ValidNumber
 *-----------------------------------------------------------------------------
 * Returns, whether a block number is valid.
 *-----------------------------------------------------------------------------
 * Input: 
 *	d0.w = block number
 * Ouput:
 *	d0.w = block number or zero on error
 *	Z-flag represents the state of d0
 */
ValidNumber:	cmpi.w	#BLOCK_NUM, d0		// block number has to be less than BLOCK_NUM
		bgt	\Error
		tst.w	d0			// block number = 0 is itself the error code
		rts
\Error:		clr.w	d0
		rts

/*-----------------------------------------------------------------------------
 * getBlockName
 *-----------------------------------------------------------------------------
 * Returns the name of the file, a block is stored in.
 *-----------------------------------------------------------------------------
 * Input: 
 *	d0.w = block number
 * Ouput:
 *	a0.l = pointer to name
 * only d0-d1 destroyed
 */
getBlockName:	lea	BlockNameNum(PC), a0	// pointer to number-part of block name
		moveq	#4-1, d1		// d1 = digit counter
		ext.l	d0			// division requires long operand
\Loop:		divu	#10, d0
		swap 	d0			// d0 = remainder
		move.b	d0, -(a0)		// copy remainder to string
		add.b	#'0', (a0)		// and make it a valid digit
		clr.w	d0			// continue conversion with the divident
		swap	d0
		dbra	d1, \Loop
		lea	BlockName(PC), a0	// return the name
		rts

/******************************************************************************
 * WriteBlock
 *-----------------------------------------------------------------------------
 * Write data pointed by a0 to the block specified by d0. The data will be
 * stored, using a simple repetition coding compression algorithm.
 *-----------------------------------------------------------------------------
 * Input: 
 *	a0.l = pointer to data
 *      d0.w = number of block ( 1..max)
 * Ouput:
 *      d0.w = zero on success, nonzero else
 *	Z-flag represents state of d0
 * Doesn't modify any registers, except d0.l
 */
//#define OLD_COMPRESSION_READ
//#define OLD_COMPRESSION_WRITE
#ifndef OLD_COMPRESSION_WRITE
forthlib@0015:
WriteBlock:
\DictSize	EQU	1031		// size of dictionary
\Probe		EQU	7		// probe offset
\Ain		EQUR	a0		// pointer to data
\Aout		EQUR	a1		// pointer to output data
\AoutStart	EQUR	a2		// beginn of output string
\AoutState	EQUR	a3		// output state (true/false) output 4+8 bit or 8+4 bit
\Atmp		EQUR	a4		// temporary reg
\Adict		EQUR	a7		// pointer to dictionary
\Dno		EQUR	d0		// block number
\Dstr		EQUR	d1		// current compression string
\Dchr		EQUR	d2		// current compression character
\Dhash		EQUR	d3		// dictionary-lookup-hash
\Dcount		EQUR	d4		// compression byte counter
\Dword		EQUR	d5		// dictionary word / dictionary lookup result
\Dcode		EQUR	d6		// next dictionary word's code
\Dtmp		EQUR	d7		// temporary reg

		movem.l	d1-d7/a0-a6, -(a7)
		bsr	ValidNumber			// d0 = zero if block number is invalid
		beq	\Error

		/** check whether whole block is zero */
		move.w	#1024/4-1, \Dcount
		movea.l	\Ain, \Atmp
\ZeroChkLoop:	tst.l	(\Atmp)+
		bne	\BlkNotZero
		dbra	\Dcount, \ZeroChkLoop
		bra	\EmptyBlock
\BlkNotZero:
		/** clear and allocate dictionary on the stack */
 		move.w	#\DictSize-1, \Dcount
\ClrDictLoop:	clr.l	-(a7)
		dbra	\Dcount, \ClrDictLoop
		
		/** initialize regs for compression */
		lea	TempBuffer, \AoutStart
		lea	2(\AoutStart), \Aout		// leave 1 word for space for var size
		suba.l	\AoutState, \AoutState		// \AoutState = 0
		clr.w	\Dstr
		clr.w	\Dchr
                clr.l	\Dhash
		move.b	(\Ain)+, \Dstr			// the first char is our first string
		move.w	#256, \Dcode			// first code-number is 256
		move.w	#1023-1, \Dcount		// count for remaining bytes to be packed

	/******* do compression ************************/
\Loop:		move.b	(\Ain)+, \Dchr			// read character

		/** search for word string+character */
		move.w	\Dstr, \Dhash			// calculate hash offset
		eor.w	\Dchr, \Dhash			//   \Dhash is our XOR hash
		lsl.w	#2, \Dhash			//   make it the byte offset
\_LookupLoop:	move.l	0(\Adict, \Dhash.w), \Dword	// bits 0..7: chr; 8..19: str; 20..31: code
		beq	\_NotFound
		cmp.b	\Dchr, \Dword			// compare (suffix) character
		bne	\__NotEqual
		move.l	\Dword, \Dtmp			// get (prefix) string code
		lsr.l	#8, \Dtmp
		andi.w	#$FFF, \Dtmp
		cmp.w	\Dstr, \Dtmp			// compare (prefix) string
		bne	\__NotEqual
		/** match found */
		swap	\Dword				// get word's code
		lsr.w	#4, \Dword		
		move.w	\Dword, \Dstr			// string = string+char
		bra	\_DoLoop			// continue loop

    		/** no match: continue at next probe position */
\__NotEqual:	add.w	\Probe*4, \Dhash		// get offset of new probe position
		ext.l	\Dhash
		divu	#\DictSize*4, \Dhash		// bound offset by modulo division
		swap	\Dhash
		bra	\_LookupLoop

		/** if we didn't found a match */
\_NotFound:	move.w	\Dcode, \Dword			// create a new word
		lsl.w	#4, \Dword			// bits 20..31 are the word's code
		lsl.l	#8, \Dword
		or.w	\Dstr, \Dword			// bits 8...20 are the prefix-string
		lsl.l	#8, \Dword
		move.b	\Dchr, \Dword			// bits 0...7 are the suffix-character
		move.l	\Dword, 0(\Adict, \Dhash.w)	// add word to empty dictionary-field
		addq.w	#1, \Dcode			// next code is current code + 1

		bsr	\OutputString			// output string now (clobbers \Dstr)
		move.w	\Dchr, \Dstr			// our new string is character

\_DoLoop:	dbra	\Dcount, \Loop			// continue loop

		bsr	\OutputString			// output code for current string

		/** remove the dictionary from the stack*/
		lea	\DictSize*4(a7), a7

		/** create STR var signature */
		clr.b	(\Aout)+		
		clr.b	(\Aout)+			// (string variable is displayed as 
		move.b	#$2D, (\Aout)+			//  empty string "" by the TIOS)
		move.w	d0, d3				// d3 = destination block number
		suba.l	\AoutStart, \Aout		// \Aout = size of compressed STR var
		move.w	a1, (\AoutStart)		// set the file-size-word at var-begin
		subq.w	#2, (\AoutStart)

		/******* copy the data to the block specified by d0 */
		move.w	\Dno, d3			// d3 = block number (isn't clobbered)
		move.l	\Aout, -(a7)			// allocate a new handle
		jsr	tios::HeapAlloc
		move.l	(a7)+, d1			// d1 = file size
		move.w	d0, d4				// d4 = d0 = handle
		beq	\Error

		bsr	DEREF_d0_a1			// a1 = pointer to memory block 
		subq.w	#1, d1				// d1 = size-1 = counter
\CopyLoop:	move.b	(\AoutStart)+, (a1)+			// copy data 
		dbra	d1, \CopyLoop
		
		lea	BlockFold(PC), a0		// create the folder for the blocks
		jsr	CreateFolder
		move.w	d3, d0				// d0 = block number
		bsr	getBlockName			// a0 = name of file where block is stored
		move.w	d4, d0				// d0 = handle of compressed STR var
		bsr	CreateFile			// create file (a0) with handle d0
		bne	\CreateError
		
\Success:	moveq	#0, d0				// d0 = success		
\End:		movem.l	(a7)+, d1-d7/a0-a6		// restore all registers
		rts

\CreateError:	move.w	d4, d0				// file creation error --> free the
		bsr	HeapFree			// compressed block
\Error:		moveq	#1, d0				// d0 = error
		bra	\End
\EmptyBlock:	bsr	getBlockName			// a0 = name of block
		bsr	DeleteFile			// delete it
		bra	\Success

/* Output \Dstr to \Aout. \Dstr is clobbered.
 */
\OutputString:	move.w	\AoutState, \Dtmp
		not.w	\Dtmp
		move.w	\Dtmp, \AoutState
		beq	\Output4_8
\Output8_4:	move.b	\Dstr, (\Aout)+
		lsr.w	#8, \Dstr
		clr.b	(\Aout)
		move.b	\Dstr, (\Aout)+
		rts
\Output4_8:	ror.w	#4, \Dstr
		move.b	\Dstr, (\Aout)+
		lsr.w	#8, \Dstr
		or.b	\Dstr, -2(\Aout)
		rts
		
/**
 * Older compression write-routine
 */
#else
forthlib@0015:
WriteBlock:	movem.l	d1-d7/a0-a6, -(a7)		// save _all_ registers

		bsr	ValidNumber			// d0 = zero if block number is invalid
		beq	\Error
	
		/******* Get the most seldom character (this is the ESC-character) */
		lea	TempBuffer, a1			//* a1 = pointer to buffer for temporary data
							//     used for storing the character frequencies
			// initialize character frequency data to zero
		move.w	#512/4-1, d7
		movea.l	a1, a2
\ClrLoop:	clr.l	(a2)+
		dbra	d7, \ClrLoop

			// calculate the frequencies
		movea.l	a0, a2				// a2 = pointer to data to scan
		move.w	#1024-1, d7			// d7 = counter
\FreqLoop:	clr.w	d1				  // d1.w is the current character
		move.b	(a2)+, d1
		lsl.w	#1, d1				  // get the offset of this character's frequency
		addq.w	#1, 0(a1, d1.w)			  // increase frequency count in the table
		dbra	d7, \FreqLoop

		cmpi.w	#1024, (a1)			// if block only consists of characters #0
		beq	\EmptyBlock			  // it is empty --> remove it

			// get the most seldom character
		moveq	#0, d1				// d1 = most seldom character
		moveq	#127, d2			// d2 = frequency of most seldom character
		moveq	#0, d7				// d7 = number of current chracter
		movea.l	a1, a2				// a2 = pointer to current character's frequency
\GetChrLoop:	move.w	(a2)+, d3			  // read character's frequency
		cmp.w	d3, d2				 
		ble	\__MoreFreq		 	  // if current char is more seldom than most seldom chr
		move.w	d7, d1				    // set the most seldom char to the current one
		move.w	d3, d2				    // set the most seldom char's frequency to the cur.
\__MoreFreq:	addq.w	#1, d7				  // go to next character
		cmpi.w	#256, d7	 		  // loop if character-number is not 256
		blt	\GetChrLoop

		/******* Compress the data to STR variable a1 (TempBuffer) */
							// d1 = ESC-character
							// a0 = pointer to source's begin
							// a1 = destination-pointer
		movea.l	a0, a2				// a2 = source-pointer
		lea	1024(a0), a3			// a3 = pointer to end of source
		move.l	a1, -(a7)			// save a1
		addq.w	#2, a1				// leave space for file size (is updated later)
		move.b	d1, (a1)+			// write ESC-character to destination
		
\ComprLoop:		// get the longest occurence of the string at the current source-address
		moveq	#1, d2				// d2 = repetition length-1
		moveq	#-1, d3				  // d3 = repetition offset
		movea.l	a0, a4			  	  // a4 = scanning-pointer
		move.b	(a2)+, d5			  // d5 = current char
\__ScanLoop:			// get the length of the repetition at the current scanning-position
		cmp.b	(a4)+, d5			    // character at the scanning-pos. = current char?
		bne	\____ContScan                       // if not: there can't be a repetition!

		moveq	#0, d4				    // d4 = current repetition's length-1
		movea.l	a4, a5				    // a5 = comparison-pointer1
		movea.l	a2, a6			    	    // a6 = comparison-pointer2
\____CmpLoop:	cmp.b	(a5)+, (a6)+			       // compare the two characters
		bne	\____EndCmp
		addq.w	#1, d4				       // if they're equal: one byte longer repet.
		cmpa.l	a2, a5			   	       // check, whether current position is reached
		bge	\____EndCmp			    
		cmpa.l	a3, a6				       // check, whether end of source is reached
		blt	\____CmpLoop
\____EndCmp:			// d4 = length of repetition
		cmp.w	d2, d4				    
		ble	\____ContScan			    // if current repetition longer than longest rep.
		move.l	a4, d3				       // set repetition offset to current position
		sub.l	a0, d3
		subq.w	#1, d3
		move.w	d4, d2				       // set repetition length to current ones
		cmpi.w	#64, d2			               // if repetition length-1 >= 64
		blt	\____ContScan
		moveq	#63, d2                                   // set it to 63
		bra	\__EndScan                                // and exit the scanning loop

\____ContScan:	cmpa.l	a2, a4				    // and loop back, if curren't position isn't reached
		blt	\__ScanLoop
\__EndScan:		// d3 = repetition offset; d2 = repetition length

		tst.w	d3				// check, whether a repetition was found
		bmi	\__NoRep

		// code a repetition
		/* Repetition Tag:  | 7 6 5 4 3 2 1 0 | 7 6  5 4 3 2 1 0 |  7  6  5  4  3  2  1  0 |
 		 *                   \ ESC-Character /  ^ ^ \Rep. Length/ \bits 2-9 of Rep. Offset/
		 *                                      bits 1 0 of Rep. Offset
		 */
		move.b	d1, (a1)+		           // write ESC-chr (byte1 of Repetition Tag )
		adda.w	d2, a2				   // set source pointer to end of repetition
		lsl.w	#6, d3				   // create byte2 of Rep.Tag
		or.b	d3, d2
		move.b	d2, (a1)+			   // write that byte
		lsr.w	#8, d3				   // create byte3 of Rep.Tag
		move.b	d3, (a1)+			   // write that byte
		bra	\__ContCompr			   // continue compression...

		// code normal character
			/* Esc. characters are coded as [ESC-Chr] [0] */
\__NoRep:	move.b	d5, (a1)+			      // write character to destination
		cmp.b	d5, d1			 	     // if it was the ESC-character
		bne	\__ContCompr
		clr.b	(a1)+				         // suspend zero

\__ContCompr:	cmpa.l	a3, a2			    	// check, whether end of source is reached
		blt	\ComprLoop                      // if it isn't: loop back

		clr.b	(a1)+				// make the buffer a string variable
		clr.b	(a1)+				// (string variable is displayed as 
		move.b	#$2D, (a1)+			//  empty string "" by the TIOS)
		move.l	(a7)+, a2			// a2 = pointer to begin of compressed data
		move.w	d0, d3				// d3 = destination block number

		/******* copy the data to the block specified by d0 */
		suba.l	a2, a1				// a1 = size of compressed STR var
		move.w	a1, (a2)			// set the file-size-word at var-begin
		subq.w	#2, (a2)
		move.l	a1, -(a7)			// allocate a new handle
		jsr	tios::HeapAlloc
		move.l	(a7)+, d1			// d1 = file size
		move.w	d0, d4				// d4 = d0 = handle
		beq	\Error

		bsr	DEREF_d0_a1			// a1 = pointer to memory block 
		subq.w	#1, d1				// d1 = size-1 = counter
\CopyLoop:	move.b	(a2)+, (a1)+			// copy data 
		dbra	d1, \CopyLoop
		
		lea	BlockFold(PC), a0		// create the folder for the blocks
		jsr	CreateFolder
		move.w	d3, d0				// d0 = block number
		bsr	getBlockName			// a0 = name of file where block is stored
		move.w	d4, d0				// d0 = handle of compressed STR var
		bsr	CreateFile			// create file (a0) with handle d0
		bne	\CreateError
		
\Success:	moveq	#0, d0				// d0 = success		
\End:		movem.l	(a7)+, d1-d7/a0-a6		// restore all registers
		rts

\CreateError:	move.w	d4, d0				// file creation error --> free the
		bsr	HeapFree			// compressed block
\Error:		moveq	#1, d0				// d0 = error
		bra	\End
\EmptyBlock:	bsr	getBlockName			// a0 = name of block
		bsr	DeleteFile			// delete it
		bra	\Success
#endif
		
/******************************************************************************
 * ReadBlock
 *-----------------------------------------------------------------------------
 * Read the data from the block, specified by d0 to the memory, addressed
 * by a0. 
 *-----------------------------------------------------------------------------
 * Input: 
 *      d0.w = number of block (1..max)
 *	a0.l = pointer to destination for uncompression
 * Ouput:
 * 	d0.w = zero on success, nonzero else
 *	Z-flag represents state of d0
 * Doesn't modify any registers!!
 */
#ifndef OLD_COMPRESSION_READ
forthlib@0016:
ReadBlock:	
\Aout		EQUR	a0
\Ain		EQUR	a1
\Adict		EQUR	a2
\Dcode		EQUR	d0
\DoldCode	EQUR	d1
\Dcount		EQUR	d2
\Dchr		EQUR	d3
\DoutCode	EQUR	d4
\DnextOffs	EQUR	d5
\DinState	EQUR	d6
\Dtmp		EQUR	d7

		movem.l	d1-d7/a0-a6, -(a7)		// save _all_ registers

		bsr	ValidNumber			// d0 = zero if block number is invalid
		beq	\Error

		/******* Get address of compressed block */
		move.l	a0, -(a7)			// save destination pointer
		bsr	getBlockName			// a0 = name of file, where blk is stored
		bsr	OpenFile			// d0 = handle or zero if not exists
		movea.l	(a7)+, a0			// restore destination pointer
		bne	\BlkNotZero		
							// if block doesn't already exist:
		move.w	#1024/4-1, d7			//   clear the block, return
\ClrLoop:	clr.l	(a0)+		
		dbra	d7, \ClrLoop
		bra	\Success

\BlkNotZero:	bsr	DEREF_d0_a1			//* a1 = pointer to compressed block
		addq.l	#2, a1				// skip file-size-field of STR var

		/** clear the dictionary (initialize fields to -1) */
		move.w	#1024-1, \Dcount
		moveq	#-1, \Dtmp
\ClearDictLoop:	move.l	\Dtmp, -(a7)
		dbra	\Dcount, \ClearDictLoop
		lea	-256*4(a7), \Adict		// dictionary entries are numbered
							// beginning from 256!

		/** intitialize regs for uncompression */
		clr.b	\DinState
		bsr	\ReadCode
		move.w	\Dcode, \DoldCode
		move.b	\Dcode, (\Aout)+
		moveq	#1, \Dcount			// counter of already outputed bytes
		move.w	#256*4, \DnextOffs

		/** uncompression... */
\ExtractLoop:	bsr	\ReadCode		
		move.w	\Dcode, \DoutCode
		bsr	\OutputString
		beq	\_Found

		/** if the string is not in the dictionary: special case */
		move.w	\DoldCode, \DoutCode
		bsr	\OutputString
		move.b	\Dchr, (\Aout)+
		addq.w	#1, \Dcount
	
\_Found:	/** add \DoldCode+\Dchr to the dictionary */
		move.w	\DoldCode, 0(\Adict, \DnextOffs)
		move.b	\Dchr, 2(\Adict, \DnextOffs)
		addq.w	#4, \DnextOffs

		move.w	\Dcode, \DoldCode
		cmpi.w	#1024, \Dcount
		blt	\ExtractLoop

		lea	1024*4(a7), a7			// remove dictionary from stack

\Success:	moveq	#0, d0				// d0 = zero -- success
		movem.l	(a7)+, d1-d7/a0-a6		// restore registers
\End:		rts
\Error:		moveq	#1, d0				// d1 = error
		bra	\End
		
/* Output the string, specified by code '\DoutCode' to '\Aout', increasing
 * '\Dcount' for every output-byte. Set '\Dchr' to the string's first character.
 * Return '\Dtmp' = zero on success. Otherwise (if it isn't in the
 * dictionary) return '\Dtmp' = nonzero.
 */
\OutputString:	cmpi.w	#256, \DoutCode			// string is a character?
		bge	\DoString
		move.b	\DoutCode, (\Aout)+		// just output it
  		move.b	\DoutCode, \Dchr		// this is also our first character
		addq.w	#1, \Dcount
		moveq	#0, \Dtmp			// return \Dtmp=0: success
		rts

\DoString:	lsl.w	#2, \DoutCode
		move.w	0(\Adict, \DoutCode), \Dtmp
		bmi	\NotInDict
		move.w	\DoutCode, -(a7)
		move.w	\Dtmp, \DoutCode		// output prefix string
		bsr	\OutputString
		move.w	(a7)+, \DoutCode

\OutputChr:	move.b	2(\Adict, \DoutCode), (\Aout)+	// output suffix character
		addq.w	#1, \Dcount
		lsr.w	#2, \DoutCode			// restore \DoutCode
  		moveq	#0, \Dtmp			// \Dtmp = 0: success
		rts

\NotInDict:	lsr.w	#2, \DoutCode
		moveq	#-1, \Dtmp
		rts
		
/* Read one code from '\Ain' to '\Dcode'.
 */
\ReadCode:	not.b	\DinState
		beq	\Read4_8
\Read8_4:	move.b	(\Ain)+, \Dtmp
		move.b	(\Ain)+, \Dcode
		andi.b	#$0F, \Dcode
		lsl.w	#8, \Dcode
		move.b	\Dtmp, \Dcode
		rts
\Read4_8:	move.b	(\Ain)+, \Dcode
		lsl.w	#8, \Dcode
		move.b	-2(\Ain), \Dcode
		lsr.w	#4, \Dcode
		rts
/**
 * Old compression's Read routine
 */
#else
forthlib@0016:
ReadBlock:	movem.l	d1-d7/a0-a6, -(a7)		// save _all_ registers

		bsr	ValidNumber			// d0 = zero if block number is invalid
		beq	\Error

		/******* Get address of compressed block */
		move.l	a0, -(a7)			// save destination pointer
		bsr	getBlockName			// a0 = name of file, where blk is stored
		bsr	OpenFile			// d0 = handle or zero if not exists
		movea.l	(a7)+, a0			// restore destination pointer
		bne	\BlkNotZero		
							// if block doesn't already exist:
		move.w	#1024/4-1, d7			//   clear the block, return
\ClrLoop:	clr.l	(a0)+		
		dbra	d7, \ClrLoop
		bra	\Success

\BlkNotZero:	bsr	DEREF_d0_a1			//* a1 = pointer to compressed block
		addq.l	#2, a1				// skip file-size-field of STR var

		/******* Uncompress the block to (a0) */
		movea.l	a0, a2				//* a2 = destination-pointer
		lea	1024(a0), a3			//* a3 = pointer to end of destination
		move.b	(a1)+, d0			//* d0 = ESC-character

\UnComprLoop:	move.b	(a1)+, d1			   // d1 = current character
		cmp.b	d1, d0				   // test, whether it is an ESC-character
		bne	\__NoESC

			// uncompress ESC-Tag
		move.b	(a1)+, d2                              // read the Repetition Tag
		beq	\__NoESC			       // if byte2 is zero: no ESC-sequence!
		clr.w	d3
		move.b	(a1)+, d3                              // d2/d3 are byte2/byte3 of tag
		move.b	d2, d4				       // d4 is the repetition's length-1
		andi.w	#$003F, d4
		lsr.b	#6, d2
		lsl.w	#2, d3
		or.b	d2, d3                                 // and d3 is the repetition's offset
		lea	0(a0, d3.w), a4                        // a4 = pointer to repetition
\__CopyLoop:	move.b	(a4)+, (a2)+                               // copy repetition to destination
		dbra	d4, \__CopyLoop
		bra	\__DoLoop

\__NoESC:		// normal character...
		move.b	d1, (a2)+				// write character to destination

\__DoLoop:	cmpa.l	a3, a2				   // if destination-end isn't reached
		blt	\UnComprLoop			   // loop back

\Success:	moveq	#0, d0				// d0 = zero -- success
		movem.l	(a7)+, d1-d7/a0-a6		// restore registers
\End:		rts
\Error:		moveq	#1, d0				// d1 = error
		bra	\End
#endif

/******************************************************************************
 ******************************************************************************
 * GRAPHIC ROUTINES
 * Note that all the following routines may only be used after `GraphicOn'
 * had been executed.
 ******************************************************************************
 ******************************************************************************/

/******************************************************************************
 * Some tools
 */
 
/**
 * Get pixel bit and offset
 *
 * Input:    d0.w = x
 *	     d1.w = y
 * Output:   if pixel is out of the screen (should be clipped):
 *              d0.l = -1; N-flag is set
 *           else:
 *	        d0.b = bit number in word (0..15)
 *              d1.w = offset of word in which the pixel is contained
 * Registers: high word of d1 is modified
 */
GetPixelOffs:	tst.w	d0		// clip pixel
		blt	\Clip
		cmpi.w	#240, d0
		bge	\Clip
		tst.w	d1
		blt	\Clip
		cmpi.w	#128, d1
		bge	\Clip

		move.w	d0, -(a7)
		lsr.w	#4, d0		// d0 = pixel's word offset from line start
		lsl.w	#1, d0
		mulu	#30, d1		// d1 = offset of line start
		add.w	d0, d1		// d1 = pixel's word offset
		move.w	(a7)+, d0
		not.w	d0		// d0 = bit number
		andi.b	#15, d0	
		rts
\Clip:		moveq	#-1, d0
		rts

/**
 * Get address of the two planes
 *
 * Input:    none
 * Output:   a0.l = pointer to least significant plane
 *           a1.l = pointer to most significant plane
 *	     Z-flag is set if grayscale is currently disabled
 */
GetPlaneAddr:	lea	Screen, a0
		movea.l	Screen.Addr-Screen(a0), a1
		move.l	Screen.Addr2-Screen(a0), -(a7)	// set Z-flag if no grayscale
		movea.l	(a7)+, a0
		rts

/**
 * Draw pixels of a mask-word in a specified color onto the grayscale-screen
 *
 * Input:    a0.l = pointer to least significant plane
 * 	     a1.l = pointer to most significant plane
 *           d0.w = word-mask
 *	     d1.w = offset of word in which the pixels should be drawn
 *	     d2.w = color
 * Output:   pixels are drawn
 * Registers: no registers are modified!
 */
DrawMask:	btst.b	#0, d2		// if color's low bit is cleared:
		beq	\BlankLSP	//  --> blank pixels in least significant plane
		or.w	d0, 0(a0, d1.w)	// else: set pixels in least significant plane
		bra	\ProcessMSP

\BlankLSP:	not.w	d0		// blank pixels inleast significant plane
		and.w	d0, 0(a0, d1.w)
		not.w	d0
		
\ProcessMSP:	btst.b	#1, d2		// if color's high bit is cleared:
		beq	\BlankMSP	//  --> blank pixels in most significant plane
		or.w	d0, 0(a1, d1.w)	// else: set pixels in least significant plane
		rts

\BlankMSP:	not.w	d0		// blank pixels inleast significant plane
		and.w	d0, 0(a1, d1.w)
		not.w	d0
		rts

/******************************************************************************
 * PutPixel
 *-----------------------------------------------------------------------------
 * Draw a grayscale pixel. If the pixel lies outside the screen, it will be
 * clipped.
 *-----------------------------------------------------------------------------
 * Input:  d0.w = x
 *         d1.w = y
 *         d2.w = color (0 = white, 1 = light grey, 2 = dark grey, 3 = black)
 * Output: Pixel is drawn
 * Note:   a0-a1/d0-d2 are destroyed! (so forgot your x, y and color arguments!)
 */
forthlib@0017:
PutPixel:	move.w	d3, -(a7)
		bsr	GetPixelOffs	// d0.w = bit number; d1.w = offset
		bmi	\End		// if N-flag is set: pixel is out of range
		bsr	GetPlaneAddr	// a0/a1 = pointer to least/most significant plane
		beq	\End		// if Z-flag is set: graphic is disabled
		moveq	#1, d3		// d3 will be our mask
		lsl.w	d0, d3
		move.w	d3, d0		// now mask is in d0
		bsr	DrawMask	// and draw the pixel
\End:		move.w	(a7)+, d3
		rts

/******************************************************************************
 * GetPixel
 *-----------------------------------------------------------------------------
 * Read a pixel and return its color.
 *-----------------------------------------------------------------------------
 * Input:  d0.w = x
 *         d1.w = y
 * Output: d2.w = color (0 = white, 1 = light grey, 2 = dark grey, 3 = black)
 *                or -1 if pixel is out of the screen
 * Note that d0-d2/a0-a1 are destroyed. (so forget your x and y arguments)
 */
forthlib@0018:
GetPixel:	bsr	GetPixelOffs	// d0.w = bit number; d1.w = offset
		bmi	\Fail		// if N-flag is set: pixel is out of range
		bsr	GetPlaneAddr	// a0/a1 = pointer to least/most significant plane
		beq	\Fail		// if Z-flag is set: graphic is disabled
		move.w  0(a0, d1.w), d2
		lsr.w	d0, d2
		andi.w	#1, d2		// d2 = least significant bit
		move.w	0(a1, d1.w), d1
		lsr.w	d0, d1		// d1 = most significant bit
		lsl.w	#1, d1
		andi.w	#2, d1
		or.w	d1, d2		// d2 = color
		rts
\Fail:		moveq	#-1, d2
		rts

/******************************************************************************
 * FillScreen
 *-----------------------------------------------------------------------------
 * Fill screen with a specified color
 *-----------------------------------------------------------------------------
 * Input:  d0.w = color (0 = white, 1 = light grey, 2 = dark grey, 3 = black)
 * Output: screen is filled with the specified color
 * Note:   a0-a1/d0-d2 are destroyed! (so forget your color argument)
 */
forthlib@0019:
FillScreen:	bsr	GetPlaneAddr	// a0/a1 = pointer to least/most significant plane
		beq	\End		// if Z-flag is set: graphic is disabled
		move.w	d0, d2		// d2 = color
		moveq	#-1, d0		// d0 = mask = all pixels!
		move.w	#30*128-2, d1	// d1 = offset of current group of pixels
\Loop:		bsr	DrawMask  	   // draw 16 pixels of mask d0 at offset d1
		subq.w	#2, d1
		bpl	\Loop
\End:		rts

/******************************************************************************
 * BlankScreen
 *-----------------------------------------------------------------------------
 * Clear the screen to color white, RAPID!
 *-----------------------------------------------------------------------------
 * Input:  none
 * Output: screen is blanked
 */
forthlib@001A:
BlankScreen:	bsr	GetPlaneAddr	// a0/a1 = pointer to least/most significant plane
		beq	\End		// if Z-flag is set: graphic is disabled
		move.w	#30*128/4-1, d0	// d0 = counter
\Loop:		clr.l	(a0)+	  	   // clear 32 pixels on least significant plane
		clr.l	(a1)+		   // clear 32 pixels on most significant plane
		dbra	d0, \Loop	
\End:		rts

		
/******************************************************************************
 ******************************************************************************
 * FLOATING POINT ROUTINES
 ******************************************************************************
 ******************************************************************************/

/******************************************************************************
 * Some tools
 */

/**
 * Return floating point representation of zero in (d0|d1)
 */
FloatZero:	moveq	#0, d0			// mantisse = 0
		move.w	#-32768, d1		// exponent = -32768
		rts

/**
 * Adjust the floating point number d0.l(mantisse) d1.w (exponent) to use all 
 * 31 significant non-sign bits.
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = resulting mantisse
 *	   d1.w = resulting exponent
 *	   d2.l is destroyed
 */
AdjustFloat:	moveq	#30, d2			// d2 = most significant bit (exclude sign bit)
		/** scan from 30 downwards until first significant bit is reached */
		tst.l	d0
		beq	\Zero
		bpl	\PosAdjLoop
\NegAdjLoop:	btst.l	d2, d0			   // scan until first used bit 
		dbeq	d2, \NegAdjLoop
		bra	\DoAdjust
\PosAdjLoop:	btst.l	d2, d0		 	   // scan until first used bit
		dbne	d2, \PosAdjLoop

\DoAdjust:	/** now d2 is the number of the first significant bit in the mantisse */
		sub.w	#30, d2			// d2 = number of bits to adjust
		neg.w	d2
		asl.l	d2, d0			// adjust mantisse
		sub.w	d2, d1			// adjust exponent
		bvs	\Zero			// return zero on overflow
		rts

\Zero:		bra	FloatZero		// return in 'FloatZero'

/******************************************************************************
 * IntToFloat
 *-----------------------------------------------------------------------------
 * Convert long interger to float.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = input (signed long int)
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 */
forthlib@001B:
IntToFloat:	moveq	#0, d1			// d1 = exponent = 0; -> number = input * 2^0
		bra	AdjustFloat		// return in 'AdjustFloat'

/******************************************************************************
 * FloatToInt
 *-----------------------------------------------------------------------------
 * Convert floating point number to integer by cutting off the fractional
 * part.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = input-mantisse
 *	   d1.w = input-exponent
 * Output: d0.l = resulting signed long integer
 */
forthlib@001C:
FloatToInt:	neg.w	d1
		cmpi.w	#31, d1
		bgt	\Zero
		tst.w	d1
		bmi	\Zero
		asr.l	d1, d0
		rts
\Zero:		moveq	#0, d0
		rts
		
/******************************************************************************
 * FloatAdd FloatSub
 *-----------------------------------------------------------------------------
 * Add/substract two floating point numbers. Each number is stored in two 
 * registers, one containing the mantisse, one containing the exponent.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse1
 *	   d1.w = exponent1
 *         d2.l = mantisse2
 *	   d3.w = exponent2
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   Returns zero on overflow
 *         no registers destroyed 
 */
/***************************************************************************/
forthlib@001D:
FloatAdd:	movem.l	d2-d4, -(a7)
		bsr	AdjustExponentForAddOrSub
	 	add.l	d2, d0			// d0 = result-mantisse
__FloatAddExp:	move.w	d4, d1			// d1 = result-exponent
		bsr	AdjustFloat		// adjust float (d0|d1)
__FloatAddRet:	movem.l	(a7)+, d2-d4
		rts

/***************************************************************************/

forthlib@001E:
FloatSub:	movem.l	d2-d4, -(a7)
		bsr	AdjustExponentForAddOrSub
		sub.l	d2, d0
		bra	__FloatAddExp
/***************************************************************************/

	/**
	 * Adjust exponents to the maximum exponent minus one, thus the resulting mantisses 
	 * will have 31 bits. Those number can in all cases be added/substracted without 
	 * overflow.
	 *
	 * Input:  (d0.l|d1.w) = float1; (d2.l|d3.w) = float2
	 * Output: d4 = new exponent
	 *	   d0.l = new mantisse of float1
	 *	   d2.l = new mantisse of float2
	 */
AdjustExponentForAddOrSub:
		/** 1.: get the maximum exponent */
	 	move.w	d1, d4			// d4 = maximum exponent
		cmp.w	d4, d3
		blt	\MaxExpGot
		move.w	d3, d4
\MaxExpGot:	addq.w	#1, d4			//      + 1
		bvc	\MaxExpOk
		bsr	FloatZero		// return zero on overflow
		move.w	d1, d4			   // new exponent = zero-exponent
		move.l	d0, d2			   // mantisse1 = mantisse2 = zero-mantisse
		rts
\MaxExpOk:
		/** 2.: convert float 1 to that exponent */
		sub.w	d4, d1			// d1 = exponent difference
		neg	d1
		cmpi.w	#31, d1			// if difference is greater than 31:
		bls	\Exp1Ok			
		moveq	#0, d0			//    insignificant mantisse, set it to zero
		bra	\AdjExp1End		// else: 
\Exp1Ok:	asr.l	d1, d0			//    adjust mantisse
\AdjExp1End:
		/** 3.: convert float 2 to that exponent */
		sub.w	d4, d3			// d3 = exponent difference
		neg	d3
		cmpi.w	#31, d3			// if difference is greater than 31:
		bls	\Exp2Ok			
		moveq	#0, d2			//    insignificant mantisse, set it to zero
		bra	\AdjExp2End		// else: 
\Exp2Ok:	asr.l	d3, d2			//    adjust mantisse
\AdjExp2End:	rts
		

/******************************************************************************
 * FloatMul FloatDiv
 *-----------------------------------------------------------------------------
 * Multiplicate/Divide two floating point numbers. Each number is stored in two
 * registers, one containing the mantisse, the other containing the exponent.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse1
 *	   d1.w = exponent1
 *         d2.l = mantisse2
 *	   d3.w = exponent2
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   Returns zero on overflow
 *         no registers destroyed 
 */
/***************************************************************************/
forthlib@001F:
FloatMul:	movem.l	d2-d7, -(a7)
	/**
	 * Calculate resulting exponent as sum of the exponents plus 32
	 */
	 	add.w	d3, d1			// d1 = product-exponent
		bvs	__FloatMulZero
		addi.w	#32, d1
		bvs	__FloatMulZero
	/**
	 * Multiplicate the two mantisses 32bit x 32bit (cross 16bit multiplication)
	 */
		/** 1.: convert to positive numbers; store resulting sign in d6*/
		bsr	ProcessSignForMulOrDiv
		
		/** 2.: get low(mantisse1) * high(mantisse2) */
	 	move.w	d0, d4			// d4 = low1
		swap	d2			// d2 = high (mantisse2)
		mulu	d2, d4			// d4 = high (low1 * high2)
		
		/** 3.: get high(mantisse1) * high(mantisse2)) */
		swap	d0			// d5 = high1
		move.w	d0, d5			//    ...
		mulu	d2, d5			// d5 = high1*high2

		/** 4.: get low(mantisse2) * high(mantisse1) */
		swap	d2			// d2 = low (mantisse2)
		move.w	d0, d7			// d7 = high (mantisse1)
		mulu	d2, d7			// d7 = high (low2 * high1)
		
		/** 5.: get insignificant high(low(mantisse1)*low(mantisse2)) for rounding */
		swap	d0			// d0 = low (mantisse1)
		mulu	d0, d2			// d2 = low(mantisse1)*low(mantisse2)
		clr.w	d2			// d2 = high (low(mantisse1)*low(mantisse2))
		swap	d2	

		/** 6.: add the results from step 2 to 4; d0 is our product */
		add.l	d7, d4			// d4 = low1*high2 + low2*high1
		moveq	#0, d3			// remember low part for rounding
		move.w	d4, d3			//   ...
		clr.w	d4			// d4.l = high (low1*high2 + low2*high1)
		swap	d4			//   ...
		move.l	d5, d0			// d0 = high1*high2 + high(low2*high1) + high(low1*high2)
		add.l	d4, d0			//    (resulting product, shifted right by 32 bit)
		
		/** 7.: increase result, using the almost insignificant part d2+d3 */
		add.l	d2, d3			// d3 = insignificant part
		moveq	#0, d4			// add insignificant part to result
		swap	d3			//   ...
		move.w	d3, d4			//   ...
		add.l	d4, d0			//   ...
		
		clr.w	d3			// check for rounding
		swap	d3
		cmpi.l	#$7FFF, d3	
		bls	\DontRoundUp
		addq.l	#1, d0
\DontRoundUp:

		/** 6.: adjust sign */
		tst.b	d6			// check negative-sign-flag
		beq	\PosResult
		neg.l	d0
\PosResult:
			
	/** 
	 * Adjust exponent and number, return 
	 */
__FloatMulAdj:	bsr	AdjustFloat
__FloatMulRet:	movem.l	(a7)+, d2-d7
		rts
__FloatMulZero:	bsr	FloatZero
		bra	__FloatMulRet

/***************************************************************************/

forthlib@0020:
FloatDiv:	movem.l	d2-d7, -(a7)
	/**
	 * Calculate resulting exponent as difference of the exponents minus 29
	 */
	 	sub.w	d3, d1			// d1 = result-exponent
		bvs	__FloatMulZero		//   ...
		subi.w	#29, d1			//   ...
		bvs	__FloatMulZero
	
	/**
	 * Divide the two mantisses to get the resulting mantisse:  
	 *    result = mantisse1 / mantisse2 * 2^29
	 */
		/** 0.: convert to positive numbers; store resulting sign in d6 */
		bsr	ProcessSignForMulOrDiv
		move.b	d6, -(a7)
		
		/** 1.: calculate high(mant1) / high(mant2) * 2^29 | remainder1 */
		move.l	d2, d4		// d4 = high (mant2)
		swap	d4		//   ...
		
		move.l	d0, d5		// d5 = high(mant1)
		clr.w	d5		//   ...
		swap	d5		//   ...
		
		divu	d4, d5		// d5.l = remainder1<<16)
		move.w	d5, d3		// d3.l = high(mant1)/high(mant2) * 2^29
		moveq	#29, d7		//   ...
		asl.l	d7, d3		//   ...
		
		/** 2.: calculate (low(mant1)+(remainder1<<16)) / high(mant2) | remainder2*/
		move.w	d0, d5		// d5 = low(mant1)+(remainder<<16)
		divu	d4, d5		// d5.l = remainder2<<16
		moveq	#0, d6		// d6.l = (low(mant1)+(remainder1<<16)) / high(mant2) * 2^(29-16)
		move.w	d5, d6		//   ...
		moveq	#13, d7		//   ...
		asl.l	d7, d6		//   ...
		
		/** 3.: calculate (remainder2<<16) / high (mant2)  / 2^3 */
		clr.w	d5		// d5 = remainder2<<16
		divu	d4, d5		// d5.w = (remainder2<<16) / high(mant2)
		moveq	#0, d7		// d7.l = (remainder2<<16) / high(mant2) / 2^3
		move.w	d5, d7		//   ...
		asr.l	#3, d7		//   ...
		
		/** 4.: calculate high(mant1)*low(mant2) / high(mant2) / 2^3 */
		swap	d0		// d0.w = high(mant1)
		mulu	d2, d0		// d0.l = high(mant1)*low(mant2)
		asr.l	#3, d0		// d0.l = high(mant1)*low(mant2) / 2^3
		divu	d4, d0		// d0.w = high(mant1)*low(mant2) / high(mant2) / 2^3
		
		/** 5.: calculate high(mant1)*low(mant2)*2^16 / (high(mant2)+96)^2 / 2^3 */
		swap	d0		// d0.w = high(mant1)*low(mant2)*2^16 / high(mant2) / 2^3
		clr.w	d0		//   ...
		divu	d4, d0		// d0.w = high(mant1)*low(mant2)*2^16 / high(mant2) / 2^3
		moveq	#0, d5		// d0.l = high(mant1)*low(mant2)*2^16 / high(mant2) / 2^3
		move.w	d0, d5		//   ...		
		
		/** 6.: add results from 1, 2, 3, substract result 5 to get resulting mantisse */
		move.l	d3, d0		// d0 = result1
		add.l	d6, d0		// d0 = result1+result2
		add.l	d7, d0		// d0 = result1+result2+result3
		sub.l	d5, d0		// d0 = result1+result2+result3 - result5

		/** 7.: adjust sign */
		move.b	(a7)+, d6	// check negative-sign-flag
		beq	\PosResult
		neg.l	d0
\PosResult:
		bra	__FloatMulAdj		// the following is same as multiplication
/***************************************************************************/
		
	/**
	 * Return whether multiplying / dividing two floats will return a negative or
	 * positive result. Convert the two floats to positive.
	 *
	 * Input:  (d0.l|d1.w) = float1    (d2.l|d3.w) = float2
	 * Output: d6.b = sign ($ff = negative; $00 = positive)
	 *         (d0.l|d1.w) = |float1|  (d2.l|d3.w) = |float2|
	 */
ProcessSignForMulOrDiv:
		sf	d6			// d6 = negative-result-flag
		tst.l	d0
		bpl	\Mant1Pos
		neg.l	d0
		st	d6
\Mant1Pos:	tst.l	d2
		bpl	\Mant2Pos
		neg.l	d2
		not.b	d6
\Mant2Pos:	rts

/******************************************************************************
 * FloatLessThan
 *-----------------------------------------------------------------------------
 * Return whether floating point number1 is less than floating point number2. 
 * Return -1 for true and 0 for false
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse1
 *	   d1.w = exponent1
 *         d2.l = mantisse2
 *	   d3.w = exponent2
 * Output: d0.w = flag (-1 = true; 0=false)
 *	   Z-flag indicates state of d0
 * Note:   no registers (except d0.l) destroyed 
 */
forthlib@0021:
FloatLessThan:
		/** first check sign and exponent */
		tst.l	d0
		bmi	\Float1Neg
		tst.l	d2		// if float1>=0 and float2<0: return false
		bmi	\False
		cmp.w	d3, d1		// check exponents
		bra	\DoExpCheck
\Float1Neg:	tst.l	d2		// if float1<0 and float2>=0: return true
		bpl	\True
		cmp.w	d1, d3		// check exponents
\DoExpCheck:	blt	\True
		bgt	\False

		/** if exponents and sign are equal: check mantisses */
		cmp.l	d2, d0		// if mantisse1<mantisse2: return true
		blt	\True
		
\False:		clr.w	d0		// false: return 0
		rts
\True:		moveq	#-1, d0		// if true: return -1
		rts

/******************************************************************************
 * FloatNegate
 *-----------------------------------------------------------------------------
 * Negate a floating point number.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   no registers destroyed 
 */
forthlib@0022
FloatNeg:
	/**
	 * Just negate the mantisse, except it is $80000000 = -2147483648, which can't
	 * be negated. In that special case the exponent has to be increased and the mantisse
	 * decreased before negation can be done.
	 */
		cmpi.l	#$80000000, d0	// mantisse = $80000000 hast to be treated special
		bne	\JustNegate	//   ...
		move.l	#$C0000000, d0	   // mantisse = mantisse/2
		addq.w	#1, d1		   // exponent = exponent+1
\JustNegate:	neg.l	d0		// negate the mantisse
		rts

/******************************************************************************
 * FloatFloor
 *-----------------------------------------------------------------------------
 * Round floating point number to an integral value using the round toward 
 * negative infinity rule. 
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 */
forthlib@0023:
FloatFloor:	tst.w	d1		// if exponent >= 0: we already have an integral value
		bpl	\Return

		cmpi.w	#-31, d1	// if exponent >= -31: clear all fractional bits
		blt	\ExpGreater31
		moveq	#-1, d2		   // d2 is our AND mask for clearing
		neg.w	d1		   // d1 = bits to the right of the comma
		lsl.l	d1, d2		   // clear those bits in the AND mask d2
		and.l	d2, d0		   // clear those bits in the mantisse
		neg.w	d1		   // restore exponent d1
		rts
		
\ExpGreater31:	tst.l	d0		// if exponent < -31: return -1 or 0 depending on sign
		bmi	\MinusOne
		bra	FloatZero	   // if mantisse >= 0: return zero
		
\MinusOne:	move.l	#$80000000, d0	   // if mantisse < 0: return -1
		move.w	#-31, d1	   //   ...
\Return:	rts

/******************************************************************************
 * FloatRound
 *-----------------------------------------------------------------------------
 * Round floating point number to an integral value using the round toward
 * nearest rule. 
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 */
forthlib@0024:
FloatRound:	movem.l	d2-d6, -(a7)
		move.l	d0, d4		// (d4|d3) = x
		move.w	d1, d3		//   ...
		bsr	FloatFloor	// (d0|d1) = floor(x)
		move.l	d4, d2		// (d2|d3) = x
		move.l	d0, d4		// (d4|d5) = floor(x)
		move.w	d1, d5		//   ...
		bsr	FloatSub	// (d0|d1) = floor(x)-x
		move.l	#$80000001, d2	// (d2|d3) = -0.499999999
		moveq	#-32, d3	//   ...
		bsr	FloatLessThan	// d0 = floor(x)-x < 0.5 ? -1 : 0
		move.w	d0, -(a7)	// push less-than flag d0
		move.l	d4, d0		// (d0|d1) = floor(x)
		move.w	d5, d1		//   ...
		tst.w	(a7)+		// if floor(x)-x <= 0.5:
		beq	\DontRoundUp	   // round up:
		move.l	#$40000000, d2	   // (d2|d3) = 1
		moveq	#-30, d3	   //   ...	
		bsr	FloatAdd	   // (d0|d1) = floor(x)+1
\DontRoundUp:	movem.l	(a7)+, d2-d6
		rts
		
		

		
		
/******************************************************************************
 * FloatLn
 *-----------------------------------------------------------------------------
 * Return the natural logarithm for a number.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   returns zero if (d0|d1) <= 0
 +         return value undefined if an overflow occures
 *         no registers destroyed 
 */
forthlib@0025:
FloatLn:	movem.l	d2-d7/a0-a6, -(a7)
	/**
	 * Input is a*2^b (2^30 <= a < 2^31). Thus our result is ln(a) + b*ln(2)
	 * Calculation of ln(a) is done in the following way:
	 *        ln(a) = integrate(differentiate(ln(a))) = integrate(1/x)
	 *    <=> ln(a) = ln(2^30) + integrate(1/x from 2^30 to a)
	 *
	 *    The integral is calculated using the Simpson approximation algorithm.
	 *    Integration boundaries are 2^30 to 2^31. 
	 *    F(0) to F(32) are our function values 1/(2^30) up to 1/a
	 *    m is the number of integration segments
	 *    h is the width of one integration segment (a-2^30)/m
	 *    So the result is: 
	 *    ln(a) = ln(2^30) + h/3 * (F(0) + F(31)
	 *                              + 2*sum(F(2i), i from 1 to m/2-1) 
	 *                              + 4*sum(F(2i-1), i from 1 to m/2))
	 *          = F(0)*h/3 + F(1)*h*4/3 + F(2)*h*2/3 + F3*h*4/3 + F(4)*h*2/3 + ... + F(32)*h/3 + ln(2^32)
	 */
		tst.l	d0		// check for negative input
		bmi	\Failure
		move.w	d1, -(a7)	// save input-exponent
		clr.w	d1		// (d0|d1) = a   (*2^b was stripped by clearing d1)
		
		/** calculate h */
		move.l	#$40000000, d2	// (d2|d3) = (a0|a1) = 2^30
		clr.w	d3		//   ...
		move.l	d2, a0		//   ...
		move.w	d3, a1		//   ...
		bsr	FloatSub	// (d0|d1) = a-2^30  (integration range)
		subq.w	#5, d1		// (d0|d1) = (a-2^30)/32  = h
		move.l	d0, a2		// (a2|a3) = h
		move.w	d1, a3		//   ...

		/** calculate h/3 */
		move.l	#$60000000, d2	// (d2|d3) = 3
		moveq	#-29, d3	//   ...
		bsr	FloatDiv	// (d0|d1) = h/3
		move.l	d0, a4		// (a4|a5) = h/3
		move.w	d1, a5		//   ...
	
		/** 
		 * calculate ln(2^30) + F(0)*h/3 + F(1)*h*4/3 + F(2)*h*2/3 + ... + F(31)*h/3 
		 * Remember:  (a0|a1) = 2^30, this is our x which is subsequently incremented by h
		 *            (a2|a3) = h, is used to increase (a0|a1) 
		 *	      (a4|a5) = h/3, is used to get F(x)*h/3, F(x)*2*h/3 and F(x)*4*h/3
		 */
		
		moveq	#0, d4		// (d4|d5) = 0  (will become our sum)
		move.w	#-32768, d5	//   ...
		moveq	#32, d7		// d7 is our counter
\IntegrateLoop:			
		/** get the divisor h*a/3 where a is 1, 2 or 4 */
		move.l	a4, d0		// (d0|d1) = h/3
		move.w	a5, d1
		tst.w	d7		// the first and last summands are F(x) = F(x) * h/3
		beq	\DivisorGot
		cmpi.w	#32, d7
		beq	\DivisorGot
		addq.w	#1, d1		// for even x: F(x) = F(x)*h*2/3 
		btst.w	#0, d7
		beq	\DivisorGot
		addq.w	#1, d1		// for odd x: F(x) = F(x)*2*2/3
\DivisorGot:
		/** calculate F(x) * h*a/3 */
		move.l	a0, d2		// (d2|d3) = x
		move.w	a1, d3		//   ...
		bsr	FloatDiv	// (d0|d1) = (h*a/3) / x = (h*a/3) * F(x)
		
		/** increase our sum (d4|d5) */
		move.l	d4, d2		// (d2|d3) = (d4|d5)
		move.w	d5, d3		//   ...
		bsr	FloatAdd	// (d0|d1) = (d2|d3) + (d0|d1)  (increased sum)
		move.l	d0, d4		// (d4|d5) = (d0|d1)
		move.w	d1, d5		//   ...
		
		/** increase our x (a0|a1) by h */
		move.l	a0, d0		// (d0|d1) = x
		move.w	a1, d1		//   ...
		move.l	a2, d2		// (d2|d3) = h
		move.l	a3, d3		//   ...
		bsr	FloatAdd	// (d0|d1) = x+h
		move.l	d0, a0		// (a0|a1) = x+h
		move.w	d1, a1		//   ...
		
		dbra	d7, \IntegrateLoop

		/**
		 * Add b * ln(2) + ln(2^30) to our result.
		 */
		move.w	(a7)+, d0	// (d0|d1) = b
		ext.l	d0		//   ...
		clr.w	d1		//   ...
		bsr	AdjustFloat	//   ...
		move.l	#1488522236, d2	// (d2|d3) = ln(2)
		moveq	#-31, d3	//   ...
		bsr	FloatMul	// (d0|d1) = b * ln(2)
		
		move.l	#1395489596, d2	// (d2|d3) = ln(2^30)
		moveq	#-26, d3	//   ...
		bsr	FloatAdd	// (d0|d1) = b * ln(2) + ln (2^30)
		
		move.l	d4, d2		// (d2|d3) = approximated integral 
		move.w	d5, d3		//   ...
		bsr	FloatAdd	// (d0|d1) = b * ln(2)

\Return:	movem.l	(a7)+, d2-d7/a0-a6
		rts
\Failure:	bsr	FloatZero
		bra	\Return
		
/******************************************************************************
 * FloatLog
 *-----------------------------------------------------------------------------
 * Return the base-ten logarithm for a number.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   returns zero if (d0|d1) <= 0
 *         return value undefined if an overflow occures
 */
forthlib@0026:
FloatLog:	/** log(a) = ln(a) / ln(10) */
		bsr	FloatLn		// (d0|d1) = ln(a)
		move.l	#1865280597, d2	// (d2|d3) = 1/ln(10)
		moveq	#-32, d3	//   ...
		move.l	d3, -(a7)
		bsr	FloatMul	// (d0|d1) = ln(a) * 1/ln(10) 
		move.l	(a7)+, d3
		rts

/******************************************************************************
 * FloatExp
 *-----------------------------------------------------------------------------
 * Return e^x for a given x.
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse
 *	   d1.w = exponent
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   return value undefined if an overflow occures
 *         no registers destroyed 
 */
forthlib@0027:
FloatExp:	movem.l	d2-d7/a0-a6, -(a7)
	/**
	 * 1. Since we are going to calculate e^x = 2^(x/ln(2)), multiplicate x with 1/ln(2)
	 */
	 	move.l	#1549082005, d2	// (d2|d3) = 1/ln(2)
		moveq	#-30, d3	//   ...
		bsr	FloatMul	// (d0|d1) = x*1/ln(2) = y

	/**
	 * 2. Split floating point number into fractional and integral part:
	 *    y = a+b | a is integral, b is real with 0 <= b < 1
	 */	
	 	move.l	d0, d4		// (d4|d5) = y
		move.w	d1, d5		//   ...
		bsr	FloatFloor	// (d0|d1) = a
		move.l	d0, a0		// (a0|d6) = a
		move.w	d1, d6		//   ...
		move.l	d4, d0		// (d0|d1) = y
		move.w	d5, d1		//   ...
		move.l	a0, d2		// (d2|d3) = a
		move.w	d6, d3		//   ...
		bsr	FloatSub	// (d0|d1) = y-a = b
		move.l	d0, a0		// (a0|d4) = b
		move.w	d1, d4		//   ...
		
		move.l	d2, d0		// (d0|d1) = a
		move.w	d3, d1		//   ...
		bsr	FloatToInt	// d0.l = a (integer)
		move.w	d0, -(a7)	// push a (is used later...)
		
	/**
	 * 3. Calculate 2^b = e^(ln(2)*b) = e^c. Since 0 <= c < 1 we can do the caculation
	 *    by using the taylor polynom e^c = c^0/0! + c^1/1! + c^2/2! + c^3/3! ...
	 */
		/** get c = ln(2)*b */
		move.l	#1488522236, d0	// (d0|d1) = ln(2)
		moveq	#-31, d1	//   ...
		move.l	a0, d2		// (d2|d3) = b
		move.w	d4, d3		//   ...
		bsr	FloatMul	// (d0|d1) = b*ln(2)
		move.l	d0, a0		// (a0|d4) = b*ln(2) = c
		move.w	d1, d4		//   ...
		
		/** initialize registers for taylor polynom */
		move.l	#$40000000, a1	// (a1|d5) = 1; this will be our b^i
		moveq	#-30, d5	//   ...
		move.l	a1, a2		// (a2|d6) = 1; this will be our i!
		move.w	d5, d6		//   ...
		suba.l	a3, a3		// (a3|a4) = 0; this will be our i
		move.w	#-32768, a4	//   ...
		
		/**
		 * To get the best accuracy we will have to add the summands of the taylor
		 * polynom in the reversed order from the smallest to the greatest. For this
		 * reason we will put the summands onto the stack, from where we read them
		 * later for addition.
		 */
		moveq	#10, d7		// d7 is our counter
\TaylorLoop:	/** get ln(2)*b^i/i! */
		move.l	a1, d0		// (d0|d1) = ln(2)*b^i
		move.w	d5, d1		//   ...
		move.l	a2, d2		// (d2|d3) = i!
		move.w	d6, d3		//   ...
		bsr	FloatDiv	// (d0|d1) = ln(2) * b^i/i!
		move.l	d0, -(a7)	// push summand (ln(2) * b^i/i!)
		move.w	d1, -(a7)	//   ...
		
		/** get ln(2)*b^(i+1) = ln(2)*b^i*b */
		move.l	a0, d0		// (d0|d1) = b
		move.w	d4, d1		//   ...
		move.l	a1, d2		// (d2|d3) = ln(2)*b^i
		move.w	d5, d3		//   ...
		bsr	FloatMul	// (d0|d1) = ln(2)*b^(i+1)
		move.l	d0, a1		// (a1|d5) = ln(2)*b^(i+1)
		move.w	d1, d5		//   ...
		
		/** get i+1 */
		move.l	a3, d0		// (d0|d1) = i
		move.w	a4, d1		//   ...
		move.l	#$40000000, d2	// (d2|d3) = 1
		moveq	#-30, d3	//   ...
		bsr	FloatAdd	// (d0|d1) = i+1
		move.l	d0, a3		// (a3|a4) = i+1
		move.w	d1, a4		//   ...

		/** get (i+1)! = i!*(i+1) */
		move.l	a2, d2		// (d2|d3) = i!
		move.w	d6, d3		//   ...
		bsr	FloatMul	// (d0|d1) = (i+1)!
		move.l	d0, a2		// (a2|d6) = (i+1)!
		move.w	d1, d6
		
		dbra	d7, \TaylorLoop
		
		/**
		 * Now take all summands from the stack and add them. 
		 */
		bsr	FloatZero	// (d0|d1) = 0; this will be our sum
		moveq	#10, d7		// d7 is our counter
\AddLoop:	move.w	(a7)+, d3	// (d2|d3) = summand
		move.l	(a7)+, d2	//   ...
		bsr	FloatAdd	// (d0|d1) = sum
		dbra	d7, \AddLoop

	/**
	 * 4. Our result will be 2^y = 2^(a+b) = 2^a * 2^b. We already have 2^b, so we need
	 *    to multiply it with 2^a, which is only a change of the result's exponent since
	 *    a is an integral value.
	 */
	 	add.w	(a7)+, d1	// (d0|d1) = (d0|d1)*2^a  (we stored on the stack, remember?)
		
		movem.l	(a7)+, d2-d7/a0-a6
		rts

/******************************************************************************
 * FloatPow
 *-----------------------------------------------------------------------------
 * Return x^y for a given x and y
 *-----------------------------------------------------------------------------
 * Input:  d0.l = mantisse x
 *	   d1.w = exponent x
 *         d2.l = mantisse y
 *         d3.w = exponent y
 * Output: d0.l = result-mantisse
 *	   d1.w = result-exponent
 * Note:   return value undefined if an overflow occures
 *         no registers destroyed 
 */
forthlib@0028:
FloatPow:
	/**
	 * calculate x^y = e^(ln(x)*y)
	 */
		bsr	FloatLn		// (d0|d1) = ln(x)
		bsr	FloatMul	// (d0|d1) = ln(x)*y
		bra	FloatExp	// (d0|d1) = e^(ln(x)*y)
					// return in 'FloatExp'


/******************************************************************************
 * DATA
 ******************************************************************************/

_library:	dc.b "forthlib", 0

Blocks:		// Data for processing the blocks
BlockName:	dc.b "blocks\blk0000"
BlockNameNum:	dc.b 0
BlockFold:	dc.b "blocks", 0

		BSS

Screen:		// Screen-data-record
Screen.Handle:	ds.w	1
Screen.Addr:	ds.l	1				// address of video-RAM
Screen.IOAddr:	ds.w	1				// (Screen.Addr/8) -- IO-port address-value
Screen.Handle2:	ds.w	1	
Screen.Addr2:	ds.l	1				// address of 2nd video-RAM for grayscale
Screen.IOAddr2:	ds.w	1				// (Screen.Addr2/8)
Screen.Phase:	ds.w	1				// flickering counter for grayscale

Cursor:		// Cursor-data-record
Cursor.ScrX	ds.w	1				// Screen-coordinates  [pixels]
Cursor.ScrY	ds.w	1
Cursor.Height	ds.w	1				// cursor height-1 
Cursor.Hidden	ds.b	1				// indicates whether cursor is hidden
Cursor.Off	ds.b	1				// indicates whether cursor is off (grayscale)
Cursor.Phase	ds.b	1				// Cursor blinking-phase (0 = off or $FF = on)
Cursor.TxtColor	ds.b	1				// color of outputed text (0 = b on w; w on b, else)
Cursor.ULflag	ds.b	1				// underline - flag
Cursor.BoldFlag	ds.b	1				// bold - flag
Cursor.Count	ds.b	1				// counter for blinking

		EVEN
		
TempBuffer:	ds.b	1024+2*4+6			// Buffer for holding temporary data (compression etc.)

		END
